import math
from abc import abstractmethod

import numpy as np

from pymoo.core.duplicate import NoDuplicateElimination
from pymoo.core.individual import Individual
from pymoo.core.infill import InfillCriterion
from pymoo.core.mixed import MixedVariableMating
from pymoo.core.parameters import get_params, flatten
from pymoo.core.problem import Problem
from pymoo.core.variable import Choice, Real, Integer, Binary
from pymoo.operators.crossover.sbx import SBX
from pymoo.operators.crossover.ux import UX
from pymoo.operators.mutation.bitflip import BFM
from pymoo.operators.mutation.pm import PM
from pymoo.operators.mutation.rm import ChoiceRandomMutation
from pymoo.operators.repair.rounding import RoundingRepair
from pymoo.operators.selection.tournament import TournamentSelection, compare


class ParameterControl:

    def __init__(self, obj) -> None:
        super().__init__()

        self.data = None

        params = flatten({"ParameterControl": get_params(obj)})
        self.params = params

        # print("PARAMETER CONTROL:", list(self.params))

    def do(self, N, set_to_params=True):
        vals = self._do(N)
        if set_to_params:
            if vals is not None:
                for k, v in vals.items():
                    self.params[k].set(v)
        return vals

    @abstractmethod
    def _do(self, N):
        pass

    def tell(self, **kwargs):
        self.data = dict(kwargs)

    def advance(self, infills=None):
        for k, v in self.params.items():
            assert len(v.get()) == len(
                infills), "Make sure that the infills and parameters asked for have the same size."
            infills.set(k, v.get())


class NoParameterControl(ParameterControl):

    def __init__(self, _) -> None:
        super().__init__(None)

    def _do(self, N):
        return {}


class RandomParameterControl(ParameterControl):

    def _do(self, N):
        return {key: value.sample(N) for key, value in self.params.items()}


class EvolutionaryParameterControl(ParameterControl):

    def __init__(self, obj) -> None:
        super().__init__(obj)
        self.eps = 0.05

    def _do(self, N):
        params = self.params
        pop = self.data.get("pop")

        # make sure that for each parameter a value exists - if not simply set it randomly
        for name, param in params.items():
            is_none = np.where(pop.get(name) == None)[0]
            if len(is_none) > 0:
                pop[is_none].set(name, param.sample(len(is_none)))

        selection = AgeBasedTournamentSelection()

        crossover = {
            Binary: UX(),
            Real: SBX(),
            Integer: SBX(vtype=float, repair=RoundingRepair()),
            Choice: UX(),
        }

        mutation = {
            Binary: BFM(),
            Real: PM(),
            Integer: PM(vtype=float, repair=RoundingRepair()),
            Choice: ChoiceRandomMutation(),
        }

        mating = MixedVariableMating(
            crossover=crossover,
            mutation=mutation,
            eliminate_duplicates=NoDuplicateElimination()
        )

        problem = Problem(vars=params)

        parents = selection(problem, pop, N, n_parents=2)
        parents = [[Individual(X={key: parent.get(key) for key in params}) for parent in mating] for mating in parents]

        off = mating(problem, parents, N, parents=True)

        Xp = off.get("X")
        ret = {param: np.array([Xp[i][param] for i in range(len(Xp))]) for param in params}

        return ret


class AgeBasedTournamentSelection(TournamentSelection):

    def __init__(self, pressure=2):
        super().__init__(age_binary_tournament, pressure)


def age_binary_tournament(pop, P, **kwargs):
    n_tournaments, n_parents = P.shape

    if n_parents != 2:
        raise ValueError("Only implemented for binary tournament!")

    S = np.full(n_tournaments, np.nan)

    for i in range(n_tournaments):
        a, b = P[i, 0], P[i, 1]
        a_gen, b_gen = pop[a].get("n_gen"), pop[b].get("n_gen")
        S[i] = compare(a, a_gen, b, b_gen, method='larger_is_better', return_random_if_equal=True)

    return S[:, None].astype(int, copy=False)


class ParameterControlMating(InfillCriterion):

    def __init__(self,
                 selection,
                 crossover,
                 mutation,
                 control=NoParameterControl,
                 **kwargs):
        super().__init__(**kwargs)
        self.selection = selection
        self.crossover = crossover
        self.mutation = mutation
        self.control = control(self)

    def _do(self, problem, pop, n_offsprings, parents=None, **kwargs):

        # how many parents need to be select for the mating - depending on number of offsprings remaining
        n_matings = math.ceil(n_offsprings / self.crossover.n_offsprings)

        # do the parameter control for the mating
        control = self.control
        control.tell(pop=pop)
        control.do(n_matings)

        # if the parents for the mating are not provided directly - usually selection will be used
        if parents is None:

            # select the parents for the mating - just an index array
            parents = self.selection.do(problem, pop, n_matings, n_parents=self.crossover.n_parents, **kwargs)

        # do the crossover using the parents index and the population - additional data provided if necessary
        off = self.crossover(problem, parents, **kwargs)

        # now we have to consider during parameter control that a crossover can produce multiple offsprings
        for name, param in control.params.items():
            param.set(np.repeat(param.get(), self.crossover.n_offsprings))

        # do the mutation on the offsprings created through crossover
        off = self.mutation(problem, off, **kwargs)

        # finally attach the parameters back to the offsprings
        control.advance(off)

        return off



from pymoo.core.termination import Termination
from pymoo.termination.delta import DeltaToleranceTermination


class ConstraintViolationTermination(DeltaToleranceTermination):

    def __init__(self, tol=1e-6, terminate_when_feasible=True, **kwargs):
        super().__init__(tol, **kwargs)
        self.terminate_when_feasible = terminate_when_feasible

    def _update(self, algorithm):
        if algorithm.problem.has_constraints():
            feasible_found = any(algorithm.opt.get("feas"))

            if feasible_found:
                if self.terminate_when_feasible:
                    return 1.0
                else:
                    return 0.0

            else:
                return super()._update(algorithm)
        else:
            return 0.0

    def _delta(self, prev, current):
        return max(0.0, prev - current)

    def _data(self, algorithm):
        return algorithm.opt.get("CV").min()


class UntilFeasibleTermination(Termination):

    def __init__(self) -> None:
        super().__init__()
        self.initial_cv = None

    def _update(self, algorithm):
        cv = algorithm.opt.get("CV").min()

        if self.initial_cv is None:
            if cv <= 0:
                self.initial_cv = 1e-32
            else:
                self.initial_cv = cv

        return 1 - cv / self.initial_cv



from copy import deepcopy

import numpy as np

from pymoo.core.parameters import get_params, flatten, set_params, hierarchical
from pymoo.core.problem import ElementwiseProblem
from pymoo.optimize import minimize


def create(algorithm, params):
    algorithm = deepcopy(algorithm)
    set_params(algorithm, hierarchical(params))
    return algorithm


class HyperparameterProblem(ElementwiseProblem):

    def __init__(self, algorithm, performance, func_create=create, vars=None, **kwargs):

        # get the parameters from the algorithm object
        if vars is None:
            vars = get_params(algorithm)

        if isinstance(vars, dict):
            vars = flatten(vars)

        assert len(vars) > 0, "No hyper-parameters found to optimize."

        super().__init__(vars=vars, **kwargs)

        self.algorithm = algorithm
        self.performance = performance
        self.func_create = func_create

    def _evaluate(self, x, out, *args, **kwargs):
        algorithm = self.func_create(self.algorithm, x)
        v = self.performance(algorithm)
        out.update(v)


class SingleObjectiveSingleRun:

    def __init__(self, problem, **kwargs):
        super().__init__()
        self.problem = problem
        self.kwargs = kwargs

    def __call__(self, algorithm):
        ret = minimize(self.problem, algorithm, **self.kwargs)
        return dict(F=ret.F, G=ret.G, H=ret.H)


def stats_single_objective_mean(rets):
    F, G, H = [], [], []
    for ret in rets:
        F.append(ret.F)
        G.append(ret.G)
        H.append(ret.H)

    F, G, H = np.array(F), np.array(G), np.array(H)

    return dict(F=F.mean(axis=0), G=G.mean(axis=0), H=H.mean(axis=0))


def stats_avg_nevals(rets):
    return dict(F=np.array([ret.algorithm.evaluator.n_eval for ret in rets]).mean())


class MultiRun:

    def __init__(self, problem, n_runs=None, seeds=None, func_stats=stats_single_objective_mean, **kwargs):
        super().__init__()
        self.problem = problem
        self.kwargs = kwargs

        if seeds is None:
            if n_runs is None:
                raise Exception("Either provide number of runs or seeds directly.")

            seeds = np.random.randint(1, 1000000, size=n_runs)

        self.seeds = seeds
        self.func_stats = func_stats

    def __call__(self, algorithm):
        rets = [minimize(self.problem, algorithm, seed=seed, **self.kwargs) for seed in self.seeds]
        out = self.func_stats(rets)
        out["__results__"] = rets
        return out



from abc import abstractmethod

from pymoo.core.operator import Operator
from pymoo.core.population import Population


class Sampling(Operator):

    def __init__(self) -> None:
        """
        This abstract class represents any sampling strategy that can be used to create an initial population or
        an initial search point.
        """
        super().__init__()

    def do(self, problem, n_samples, **kwargs):
        """
        Sample new points with problem information if necessary.

        Parameters
        ----------

        problem : :class:`~pymoo.core.problem.Problem`
            The problem to which points should be sampled. (lower and upper bounds, discrete, binary, ...)

        n_samples : int
            Number of samples

        Returns
        -------
        pop : Population
            The output population after sampling

        """
        val = self._do(problem, n_samples, **kwargs)
        return Population.new("X", val)

    @abstractmethod
    def _do(self, problem, n_samples, **kwargs):
        pass







import numpy as np

import pymoo.gradient.toolbox as anp
from pymoo.core.problem import Problem
from pymoo.util.normalization import normalize


class ZDT(Problem):

    def __init__(self, n_var=30, **kwargs):
        super().__init__(n_var=n_var, n_obj=2, xl=0, xu=1, vtype=float, **kwargs)


class ZDT1(ZDT):

    def _calc_pareto_front(self, n_pareto_points=100):
        x = np.linspace(0, 1, n_pareto_points)
        return np.array([x, 1 - np.sqrt(x)]).T

    def _evaluate(self, x, out, *args, **kwargs):
        f1 = x[:, 0]
        g = 1 + 9.0 / (self.n_var - 1) * anp.sum(x[:, 1:], axis=1)
        f2 = g * (1 - anp.power((f1 / g), 0.5))

        out["F"] = anp.column_stack([f1, f2])


class ZDT2(ZDT):

    def _calc_pareto_front(self, n_pareto_points=100):
        x = np.linspace(0, 1, n_pareto_points)
        return np.array([x, 1 - np.power(x, 2)]).T

    def _evaluate(self, x, out, *args, **kwargs):
        f1 = x[:, 0]
        c = anp.sum(x[:, 1:], axis=1)
        g = 1.0 + 9.0 * c / (self.n_var - 1)
        f2 = g * (1 - anp.power((f1 * 1.0 / g), 2))

        out["F"] = anp.column_stack([f1, f2])


class ZDT3(ZDT):

    def _calc_pareto_front(self, n_points=100, flatten=True):
        regions = [[0, 0.0830015349],
                   [0.182228780, 0.2577623634],
                   [0.4093136748, 0.4538821041],
                   [0.6183967944, 0.6525117038],
                   [0.8233317983, 0.8518328654]]

        pf = []

        for r in regions:
            x1 = np.linspace(r[0], r[1], int(n_points / len(regions)))
            x2 = 1 - np.sqrt(x1) - x1 * np.sin(10 * np.pi * x1)
            pf.append(np.array([x1, x2]).T)

        if not flatten:
            pf = np.concatenate([pf[None,...] for pf in pf])
        else:
            pf = np.row_stack(pf)

        return pf

    def _evaluate(self, x, out, *args, **kwargs):
        f1 = x[:, 0]
        c = anp.sum(x[:, 1:], axis=1)
        g = 1.0 + 9.0 * c / (self.n_var - 1)
        f2 = g * (1 - anp.power(f1 * 1.0 / g, 0.5) - (f1 * 1.0 / g) * anp.sin(10 * anp.pi * f1))

        out["F"] = anp.column_stack([f1, f2])


class ZDT4(ZDT):
    def __init__(self, n_var=10):
        super().__init__(n_var)
        self.xl = -5 * np.ones(self.n_var)
        self.xl[0] = 0.0
        self.xu = 5 * np.ones(self.n_var)
        self.xu[0] = 1.0
        self.func = self._evaluate

    def _calc_pareto_front(self, n_pareto_points=100):
        x = np.linspace(0, 1, n_pareto_points)
        return np.array([x, 1 - np.sqrt(x)]).T

    def _evaluate(self, x, out, *args, **kwargs):
        f1 = x[:, 0]
        g = 1.0
        g += 10 * (self.n_var - 1)
        for i in range(1, self.n_var):
            g += x[:, i] * x[:, i] - 10.0 * anp.cos(4.0 * anp.pi * x[:, i])
        h = 1.0 - anp.sqrt(f1 / g)
        f2 = g * h

        out["F"] = anp.column_stack([f1, f2])


class ZDT5(ZDT):

    def __init__(self, m=11, n=5, normalize=True, **kwargs):
        self.m = m
        self.n = n
        self.normalize = normalize
        super().__init__(n_var=(30 + n * (m - 1)), **kwargs)

    def _calc_pareto_front(self, n_pareto_points=100):
        x = 1 + np.linspace(0, 1, n_pareto_points) * 30
        pf = np.column_stack([x, (self.m-1) / x])
        if self.normalize:
            pf = normalize(pf)
        return pf

    def _evaluate(self, x, out, *args, **kwargs):
        x = x.astype(float)

        _x = [x[:, :30]]
        for i in range(self.m - 1):
            _x.append(x[:, 30 + i * self.n: 30 + (i + 1) * self.n])

        u = anp.column_stack([x_i.sum(axis=1) for x_i in _x])
        v = (2 + u) * (u < self.n) + 1 * (u == self.n)
        g = v[:, 1:].sum(axis=1)

        f1 = 1 + u[:, 0]
        f2 = g * (1 / f1)

        if self.normalize:
            f1 = normalize(f1, 1, 31)
            f2 = normalize(f2, (self.m-1) * 1/31, (self.m-1))

        out["F"] = anp.column_stack([f1, f2])


class ZDT6(ZDT):

    def __init__(self, n_var=10, **kwargs):
        super().__init__(n_var=n_var, **kwargs)

    def _calc_pareto_front(self, n_pareto_points=100):
        x = np.linspace(0.2807753191, 1, n_pareto_points)
        return np.array([x, 1 - np.power(x, 2)]).T

    def _evaluate(self, x, out, *args, **kwargs):
        f1 = 1 - anp.exp(-4 * x[:, 0]) * anp.power(anp.sin(6 * anp.pi * x[:, 0]), 6)
        g = 1 + 9.0 * anp.power(anp.sum(x[:, 1:], axis=1) / (self.n_var - 1.0), 0.25)
        f2 = g * (1 - anp.power(f1 / g, 2))

        out["F"] = anp.column_stack([f1, f2])



import numpy as np

from pymoo.decomposition.asf import ASF


class AASF(ASF):

    def __init__(self, eps=1e-10, _type="auto", rho=None, beta=None, **kwargs) -> None:
        super().__init__(eps, _type, **kwargs)
        if rho is None and beta is None:
            raise Exception("Either provide rho or beta!")
        elif rho is None:
            self.rho = calc_rho(beta)
        else:
            self.rho = rho

    def _do(self, F, weights, **kwargs):
        asf = super()._do(F, weights, **kwargs)
        augment = ((F - self.utopian_point) / np.clip(weights, 1e-12, np.inf)).sum(axis=1)
        return asf + self.rho * augment


def calc_rho(beta):
    return 1 / (1 - np.tan(beta / 360 * 2 * np.pi)) - 1



from pymoo.core.termination import Termination


class MaximumGenerationTermination(Termination):

    def __init__(self, n_max_gen=float("inf")) -> None:
        super().__init__()
        self.n_max_gen = n_max_gen

    def _update(self, algorithm):
        if self.n_max_gen is None:
            return 0.0
        else:
            return algorithm.n_gen / self.n_max_gen




import importlib

from pymoo.config import Config


def get_functions():
    from pymoo.util.nds.fast_non_dominated_sort import fast_non_dominated_sort
    from pymoo.util.nds.efficient_non_dominated_sort import efficient_non_dominated_sort
    from pymoo.util.nds.tree_based_non_dominated_sort import tree_based_non_dominated_sort
    from pymoo.decomposition.util import calc_distance_to_weights
    from pymoo.util.misc import calc_perpendicular_distance
    from pymoo.util.hv import hv
    from pymoo.util.stochastic_ranking import stochastic_ranking

    FUNCTIONS = {
        "fast_non_dominated_sort": {
            "python": fast_non_dominated_sort, "cython": "pymoo.cython.non_dominated_sorting"
        },
        "efficient_non_dominated_sort": {
            "python": efficient_non_dominated_sort, "cython": "pymoo.cython.non_dominated_sorting"
        },
        "tree_based_non_dominated_sort": {
            "python": tree_based_non_dominated_sort, "cython": "pymoo.cython.non_dominated_sorting"
        },
        "calc_distance_to_weights": {
            "python": calc_distance_to_weights, "cython": "pymoo.cython.decomposition"
        },
        "calc_perpendicular_distance": {
            "python": calc_perpendicular_distance, "cython": "pymoo.cython.calc_perpendicular_distance"
        },
        "stochastic_ranking": {
            "python": stochastic_ranking, "cython": "pymoo.cython.stochastic_ranking"
        },
        "hv": {
            "python": hv, "cython": "pymoo.cython.hv"
        },

    }

    return FUNCTIONS


class FunctionLoader:
    # -------------------------------------------------
    # Singleton Pattern
    # -------------------------------------------------
    __instance = None

    @staticmethod
    def get_instance():
        if FunctionLoader.__instance is None:
            FunctionLoader.__instance = FunctionLoader()
        return FunctionLoader.__instance

    # -------------------------------------------------

    def __init__(self) -> None:
        super().__init__()
        self.is_compiled = is_compiled()
        self.mode = "auto"

        if Config.warnings["not_compiled"] and not self.is_compiled:
            print("\nCompiled modules for significant speedup can not be used!")
            print("https://pymoo.org/installation.html#installation")
            print()
            print("To disable this warning:")
            print("from pymoo.config import Config")
            print("Config.warnings['not_compiled'] = False\n")

    def load(self, func_name=None, mode=None):
        if mode is None:
            mode = self.mode

        FUNCTIONS = get_functions()

        if mode == "auto":
            mode = "cython" if self.is_compiled else "python"

        if func_name not in FUNCTIONS:
            raise Exception("Function %s not found: %s" % (func_name, FUNCTIONS.keys()))

        func = FUNCTIONS[func_name]
        if mode not in func:
            raise Exception("Module not available in %s." % mode)
        func = func[mode]

        # either provide a function or a string to the module (used for cython)
        if not callable(func):
            module = importlib.import_module(func)
            func = getattr(module, func_name)

        return func


def load_function(func_name=None, _type="auto"):
    return FunctionLoader.get_instance().load(func_name, mode=_type)


def is_compiled():
    try:
        from pymoo.cython.info import info
        if info() == "yes":
            return True
        else:
            return False
    except:
        return False



"""
NOTE: The factory methods have been deprecated since version 0.5.0.
The reason is the additional maintenance and the fact that the individual constructor parameters are kept hidden.

Only get_termination and get_problem are still active and are now implemented in the corresponding modules.
They will be routed through this factory method.

"""

from deprecated import deprecated


@deprecated("Please use `from pymoo.termination import get_termination`")
def get_termination(name, *args, **kwargs):
    from pymoo.termination import get_termination as func
    return func(name, *args, **kwargs)


@deprecated("Please use `from pymoo.util.ref_dirs import get_reference_directions`")
def get_reference_directions(name, *args, **kwargs):
    from pymoo.util.ref_dirs import get_reference_directions as func
    return func(name, *args, **kwargs)


@deprecated("Please use `from pymoo.problems import get_problem\n`")
def get_problem(name, *args, **kwargs):
    from pymoo.problems import get_problem as func
    return func(name, *args, **kwargs)


def get_algorithm(*args, **kwargs):
    raise Exception("The method `get_algorithm` has been deprecated since 0.6.0\n"
                    "Please use the object-oriented interface.")


def get_sampling(*args, **kwargs):
    raise Exception("The method `get_sampling` has been deprecated since 0.6.0\n"
                    "Please use the object-oriented interface:\n"
                    "https://pymoo.org/operators/selection.html")


def get_selection(*args, **kwargs):
    raise Exception("The method `get_selection` has been deprecated since 0.6.0\n"
                    "Please use the object-oriented interface:\n"
                    "https://pymoo.org/operators/selection.html")


def get_crossover(*args, **kwargs):
    raise Exception("The method `get_crossover` has been deprecated since 0.6.0\n"
                    "Please use the object-oriented interface:\n"
                    "https://pymoo.org/operators/crossover.html")


def get_mutation(*args, **kwargs):
    raise Exception("The method `get_mutation` has been deprecated since 0.6.0\n"
                    "Please use the object-oriented interface:\n"
                    "https://pymoo.org/operators/mutation.html")


def get_visualization(*args, **kwargs):
    raise Exception("The method `get_visualization` has been deprecated since 0.6.0\n"
                    "Please use the object-oriented interface:\n"
                    "https://pymoo.org/visualization/index.html")


def get_performance_indicator(*args, **kwargs):
    raise Exception("The method `get_performance_indicator` has been deprecated since 0.6.0\n"
                    "Please use the object-oriented interface:\n"
                    "https://pymoo.org/misc/indicators.html")


def get_decision_making(*args, **kwargs):
    raise Exception("The method `get_decision_making` has been deprecated since 0.6.0\n"
                    "Please use the object-oriented interface.")



from pymoo.algorithms.soo.nonconvex.ga import GA
from pymoo.core.problem import Problem
from pymoo.optimize import minimize
from pymoo.util.reference_direction import get_partition_closest_to_points, ReferenceDirectionFactory


class ReferenceDirectionGA(ReferenceDirectionFactory):

    def __init__(self,
                 n_dim,
                 n_points,
                 fun,
                 pop_size=20,
                 n_gen=200,
                 verbose=False,
                 **kwargs):

        super().__init__(n_dim, **kwargs)

        self.n_points = n_points
        self.pop_size = pop_size
        self.n_gen = n_gen

        self.fun = fun
        self.verbose = verbose

    def _do(self):
        pop_size, n_gen = self.pop_size, self.n_gen
        n_points, n_dim, = self.n_points, self.n_dim
        fun = self.fun

        class MyProblem(Problem):

            def __init__(self):
                self.n_points = n_points
                self.n_dim = n_dim
                self.n_partitions = get_partition_closest_to_points(n_points, n_dim)

                super().__init__(n_var=n_points * n_dim,
                                 n_obj=1,
                                 xl=0.0,
                                 xu=1.0,
                                 elementwise_evaluation=True)

            def get_points(self, x):
                _x = x.reshape((self.n_points, self.n_dim)) ** 2
                _x = _x / _x.sum(axis=1)[:, None]
                return _x

            def _evaluate(self, x, out, *args, **kwargs):
                out["F"] = fun(self.get_points(x))

        problem = MyProblem()

        algorithm = GA(pop_size=pop_size, eliminate_duplicates=True)

        res = minimize(problem,
                       algorithm,
                       termination=('n_gen', n_gen),
                       verbose=True)

        ref_dirs = problem.get_points(res.X)
        return ref_dirs



import numpy as np

from pymoo.util.misc import swap


def randomized_argsort(A, method="numpy", order='ascending'):
    if method == "numpy":
        P = np.random.permutation(len(A))
        I = np.argsort(A[P], kind='quicksort')
        I = P[I]

    elif method == "quicksort":
        I = quicksort(A)

    else:
        raise Exception("Randomized sort method not known.")

    if order == 'ascending':
        return I
    elif order == 'descending':
        return np.flip(I, axis=0)
    else:
        raise Exception("Unknown sorting order: ascending or descending.")


def quicksort(A):
    I = np.arange(len(A))
    _quicksort(A, I, 0, len(A) - 1)
    return I


def _quicksort(A, I, left, right):
    if left < right:

        index = np.random.randint(left, right + 1)
        swap(I, right, index)

        pivot = A[I[right]]

        i = left - 1

        for j in range(left, right):

            if A[I[j]] <= pivot:
                i += 1
                swap(I, i, j)

        index = i + 1
        swap(I, right, index)

        _quicksort(A, I, left, index - 1)
        _quicksort(A, I, index + 1, right)


if __name__ == '__main__':
    a = np.array([5, 9, 10, 0, 0, 0, 100, -2])

    for i in range(200):
        I = randomized_argsort(a, method="numpy")
        print(I)



import cma
import numpy as np

from pymoo.algorithms.base.local import LocalSearch
from pymoo.core.population import Population
from pymoo.core.termination import NoTermination
from pymoo.docs import parse_doc_string
from pymoo.termination.max_eval import MaximumFunctionCallTermination
from pymoo.termination.max_gen import MaximumGenerationTermination
from pymoo.util.display.column import Column
from pymoo.util.display.single import SingleObjectiveOutput
from pymoo.util.normalization import ZeroToOneNormalization, NoNormalization
from pymoo.util.optimum import filter_optimum
from pymoo.vendor.vendor_cmaes import my_fmin


# =========================================================================================================
# Implementation
# =========================================================================================================


class CMAESOutput(SingleObjectiveOutput):

    def __init__(self):
        super().__init__()

        self.sigma = Column("sigma")
        self.min_std = Column("min_std", width=8)
        self.max_std = Column("max_std", width=8)
        self.axis = Column("axis", width=8)

        self.run = Column("run", width=4)
        self.fpop = Column("fpop", width=8)
        self.n_pop = Column("n_pop", width=5)

    def initialize(self, algorithm):
        super().initialize(algorithm)

        if algorithm.restarts > 0:
            self.columns += [self.run, self.fpop, self.n_pop]

        self.columns += [self.sigma, self.min_std, self.max_std, self.axis]

    def update(self, algorithm):
        super().update(algorithm)

        if not algorithm.es.gi_frame:
            return

        fmin = algorithm.es.gi_frame.f_locals
        cma = fmin["es"]

        self.sigma.set(cma.sigma)

        val = cma.sigma_vec * cma.dC ** 0.5
        self.min_std.set((cma.sigma * min(val)))
        self.max_std.set((cma.sigma * max(val)))

        if algorithm.restarts > 0:
            self.run.set(int(fmin["irun"] - fmin["runs_with_small"]) + 1)
            self.fpop.set(algorithm.pop.get("F").min())
            self.n_pop.set(int(cma.opts['popsize']))

        axis = (cma.D.max() / cma.D.min()
                if not cma.opts['CMA_diagonal'] or cma.countiter > cma.opts['CMA_diagonal']
                else max(cma.sigma_vec * 1) / min(cma.sigma_vec * 1))
        self.axis.set(axis)


class CMAES(LocalSearch):

    def __init__(self,
                 x0=None,
                 sigma=0.1,
                 normalize=True,
                 parallelize=True,
                 maxfevals=np.inf,
                 tolfun=1e-11,
                 tolx=1e-11,
                 restarts=0,
                 restart_from_best='False',
                 incpopsize=2,
                 eval_initial_x=False,
                 noise_handler=None,
                 noise_change_sigma_exponent=1,
                 noise_kappa_exponent=0,
                 bipop=False,
                 cmaes_verbose=-9,
                 verb_log=0,
                 output=CMAESOutput(),
                 **kwargs
                 ):
        """


        Parameters
        ----------

        x0 : list or `numpy.ndarray`
              initial guess of minimum solution
              before the application of the geno-phenotype transformation
              according to the ``transformation`` option.  It can also be
              a string holding a Python expression that is evaluated
              to yield the initial guess - this is important in case
              restarts are performed so that they start from different
              places.  Otherwise ``x0`` can also be a `cma.CMAEvolutionStrategy`
              object instance, in that case ``sigma0`` can be ``None``.

        sigma : float
              Initial standard deviation in each coordinate.
              ``sigma0`` should be about 1/4th of the search domain width
              (where the optimum is to be expected). The variables in
              ``objective_function`` should be scaled such that they
              presumably have similar sensitivity.
              See also `ScaleCoordinates`.

        parallelize : bool
              Whether the objective function should be called for each single evaluation or batch wise.

        restarts : int, default 0
              Number of restarts with increasing population size, see also
              parameter ``incpopsize``, implementing the IPOP-CMA-ES restart
              strategy, see also parameter ``bipop``; to restart from
              different points (recommended), pass ``x0`` as a string.

        restart_from_best : bool, default false
               Which point to restart from

        incpopsize : int
              Multiplier for increasing the population size ``popsize`` before each restart

        eval_initial_x : bool
              Evaluate initial solution, for ``None`` only with elitist option

        noise_handler : class
              A ``NoiseHandler`` class or instance or ``None``. Example:
              ``cma.fmin(f, 6 * [1], 1, noise_handler=cma.NoiseHandler(6))``
              see ``help(cma.NoiseHandler)``.

        noise_change_sigma_exponent : int
              Exponent for the sigma increment provided by the noise handler for
              additional noise treatment. 0 means no sigma change.

        noise_kappa_exponent : int
              Instead of applying reevaluations, the "number of evaluations"
              is (ab)used as init_simplex_scale factor kappa (experimental).

        bipop : bool
              If `True`, run as BIPOP-CMA-ES; BIPOP is a special restart
              strategy switching between two population sizings - small
              (like the default CMA, but with more focused search) and
              large (progressively increased as in IPOP). This makes the
              algorithm perform well both on functions with many regularly
              or irregularly arranged local optima (the latter by frequently
              restarting with small populations).  For the `bipop` parameter
              to actually take effect, also select non-zero number of
              (IPOP) restarts; the recommended setting is ``restarts<=9``
              and `x0` passed as a string using `numpy.rand` to generate
              initial solutions. Note that small-population restarts
              do not count into the total restart count.

        AdaptSigma : True
              Or False or any CMAAdaptSigmaBase class e.g. CMAAdaptSigmaTPA, CMAAdaptSigmaCSA

        CMA_active : True
              Negative update, conducted after the original update

        CMA_activefac : 1
              Learning rate multiplier for active update

        CMA_cmean : 1
              Learning rate for the mean value

        CMA_const_trace : False
            Normalize trace, 1, True, "arithm", "geom", "aeig", "geig" are valid

        CMA_diagonal : 0*100*N/popsize**0.5
            Number of iterations with diagonal covariance matrix, True for always

        CMA_eigenmethod : np.linalg.eigh or cma.utilities.math.eig or pygsl.eigen.eigenvectors

        CMA_elitist : False  or "initial" or True
            Elitism likely impairs global search performance

        CMA_injections_threshold_keep_len : 0
            Keep length if Mahalanobis length is below the given relative threshold

        CMA_mirrors : popsize < 6
            Values <0.5 are interpreted as fraction, values >1 as numbers (rounded), otherwise about 0.16 is used

        CMA_mirrormethod : int, default 2,  0=unconditional, 1=selective, 2=selective with delay

        CMA_mu : None
            Parents selection parameter, default is popsize // 2

        CMA_on : 1
            Multiplier for all covariance matrix updates

        CMA_sampler : None
            A class or instance that implements the interface of
              `cma.interfaces.StatisticalModelSamplerWithZeroMeanBaseClass`

        CMA_sampler_options : dict
            Options passed to `CMA_sampler` class init as keyword arguments

        CMA_rankmu : 1.0
            Multiplier for rank-mu update learning rate of covariance matrix

        CMA_rankone : 1.0
            Multiplier for rank-one update learning rate of covariance matrix

        CMA_recombination_weights : None
            A list, see class RecombinationWeights, overwrites CMA_mu and popsize options

        CMA_dampsvec_fac : np.Inf
            Tentative and subject to changes, 0.5 would be a "default" damping for sigma vector update

        CMA_dampsvec_fade : 0.1
            Tentative fading out parameter for sigma vector update

        CMA_teststds : None
            Factors for non-isotropic initial distr. of C, mainly for test purpose, see CMA_stds for production

        CMA_stds : None
            Multipliers for sigma0 in each coordinate, not represented in C,
            makes scaling_of_variables obsolete

        CSA_dampfac : 1
            Positive multiplier for step-size damping, 0.3 is close to optimal on the sphere

        CSA_damp_mueff_exponent : 0.5
            Zero would mean no dependency of damping on mueff, useful with CSA_disregard_length option

        CSA_disregard_length : False
            True is untested, also changes respective parameters

        CSA_clip_length_value : None
            Poorly tested, [0, 0] means const length N**0.5, [-1, 1] allows a variation of +- N/(N+2), etc.

        CSA_squared : False
            Use squared length for sigma-adaptation ',

        BoundaryHandler : BoundTransform or BoundPenalty, unused when ``bounds in (None, [None, None])``

        conditioncov_alleviate : [1e8, 1e12]
            When to alleviate the condition in the coordinates and in main axes

        eval_final_mean : True
            Evaluate the final mean, which is a favorite return candidate

        fixed_variables : None
            Dictionary with index-value pairs like dict(0=1.1, 2=0.1) that are not optimized

        ftarget : -inf
            Target function value, minimization

        integer_variables : []
            Index list, invokes basic integer handling: prevent std dev to become too small in the given variables

        maxfevals : inf
            Maximum number of function evaluations

        maxiter : 100 + 150 * (N+3)**2 // popsize**0.5
            Maximum number of iterations

        mean_shift_line_samples : False
            Sample two new solutions colinear to previous mean shift

        mindx : 0
            Minimal std in any arbitrary direction, cave interference with tol

        minstd : 0
            Minimal std (scalar or vector) in any coordinate direction, cave interference with tol

        maxstd : inf
            Maximal std in any coordinate direction

        pc_line_samples : False
            One line sample along the evolution path pc

        popsize : 4+int(3*np.log(N))
            Population size, AKA lambda, number of new solution per iteration

        randn : np.random.randn
            Randn(lam, N) must return an np.array of shape (lam, N), see also cma.utilities.math.randhss

        signals_filename : None
            cma_signals.in  # read versatile options from this file which contains a single options dict,
            e.g. ``dict("timeout"=0)`` to stop, string-values are evaluated, e.g. "np.inf" is valid

        termination_callback : None
            A function returning True for termination, called in `stop` with `self` as argument, could be abused
            for side effects

        timeout : inf
            Stop if timeout seconds are exceeded, the string "2.5 * 60**2" evaluates to 2 hours and 30 minutes

        tolconditioncov : 1e14
            Stop if the condition of the covariance matrix is above `tolconditioncov`

        tolfacupx : 1e3
            Termination when step-size increases by tolfacupx (diverges). That is, the initial step-size was chosen
            far too small and better solutions were found far away from the initial solution x0

        tolupsigma : 1e20
            Sigma/sigma0 > tolupsigma * max(eivenvals(C)**0.5) indicates "creeping behavior" with usually minor
            improvements

        tolfun : 1e-11
            Termination criterion: tolerance in function value, quite useful

        tolfunhist : 1e-12
            Termination criterion: tolerance in function value history

        tolstagnation : int(100 + 100 * N**1.5 / popsize)
            Termination if no improvement over tolstagnation iterations

        tolx : 1e-11
            Termination criterion: tolerance in x-changes

        typical_x : None
            Used with scaling_of_variables',

        updatecovwait : None
            Number of iterations without distribution update, name is subject to future changes

        cmaes_verbose : 3
            Verbosity e.g. of initial/final message, -1 is very quiet, -9 maximally quiet, may not be fully implemented

        verb_append : 0
            Initial evaluation counter, if append, do not overwrite output files

        verb_disp : 100
            Verbosity: display console output every verb_disp iteration

        verb_filenameprefix : str
            CMADataLogger.default_prefix + Output path and filenames prefix

        verb_log : 1
            Verbosity: write data to files every verb_log iteration, writing can be time critical on fast to
            evaluate functions

        verb_plot : 0
              In fmin(): plot() is called every verb_plot iteration

        verb_time : True
              Output timings on console

        vv : dict
            Versatile set or dictionary for hacking purposes, value found in self.opts["vv"]

        kwargs : dict
              A dictionary with additional options passed to the constructor
              of class ``CMAEvolutionStrategy``, see ``cma.CMAOptions`` ()
              for a list of available options.

        """
        super().__init__(x0=x0, output=output, **kwargs)

        self.termination = NoTermination()

        self.es = None
        self.cma = None

        self.normalize = normalize
        self.norm = None

        self.sigma = sigma
        self.restarts = restarts
        self.restart_from_best = restart_from_best
        self.incpopsize = incpopsize
        self.eval_initial_x = eval_initial_x
        self.noise_handler = noise_handler
        self.noise_change_sigma_exponent = noise_change_sigma_exponent
        self.noise_kappa_exponent = noise_kappa_exponent
        self.bipop = bipop

        self.options = dict(
            verbose=cmaes_verbose,
            verb_log=verb_log,
            maxfevals=maxfevals,
            tolfun=tolfun,
            tolx=tolx,
            **kwargs
        )

        self.send_array_to_yield = True
        self.parallelize = parallelize
        self.al = None

    def _setup(self, problem, **kwargs):

        xl, xu = problem.bounds()
        if self.normalize:
            self.norm, self.options['bounds'] = bounds_if_normalize(xl, xu)
        else:
            self.norm = NoNormalization()
            self.options['bounds'] = [xl, xu]

        self.options['seed'] = self.seed

        if isinstance(self.termination, MaximumGenerationTermination):
            self.options['maxiter'] = self.termination.n_max_gen
        elif isinstance(self.termination, MaximumFunctionCallTermination):
            self.options['maxfevals'] = self.termination.n_max_evals

    def _initialize_advance(self, **kwargs):
        super()._initialize_advance(**kwargs)

        kwargs = dict(
            options=self.options,
            parallel_objective=self.parallelize,
            restarts=self.restarts,
            restart_from_best=self.restart_from_best,
            incpopsize=self.incpopsize,
            eval_initial_x=self.eval_initial_x,
            noise_handler=self.noise_handler,
            noise_change_sigma_exponent=self.noise_change_sigma_exponent,
            noise_kappa_exponent=self.noise_kappa_exponent,
            bipop=self.bipop)

        x0 = self.norm.forward(self.x0.X)
        self.es = my_fmin(x0, self.sigma, **kwargs)

        # do this to allow the printout in the first generation
        self.next_X = next(self.es)

    def _infill(self):

        X = np.array(self.next_X)
        self.send_array_to_yield = X.ndim > 1
        X = np.atleast_2d(X)

        # evaluate the population
        self.pop = Population.new("X", self.norm.backward(X))

        return self.pop

    def _advance(self, infills=None, **kwargs):

        if infills is None:
            self.termination.force_termination = True

        else:

            # set infeasible individual's objective values to np.nan - then CMAES can handle it
            for ind in infills:
                if not ind.feas:
                    ind.F[:] = np.nan

            F = infills.get("f").tolist()
            if not self.send_array_to_yield:
                F = F[0]

            try:
                self.next_X = self.es.send(F)
            except:
                self.next_X = None

            if self.next_X is None:
                self.termination.force_termination = True

    def _set_optimum(self):
        pop = self.pop if self.opt is None else Population.merge(self.opt, self.pop)
        self.opt = filter_optimum(pop, least_infeasible=True)

    def __getstate__(self):
        state = self.__dict__.copy()
        state.pop("es", None)
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.ers = None


class SimpleCMAES(LocalSearch):

    def __init__(self, sigma=0.1, opts=None, normalize=True, **kwargs):
        super().__init__(**kwargs)
        self.termination = NoTermination()
        self.es = None
        self.sigma = sigma
        self.normalize = normalize
        self.norm = None

        DEFAULTS = {"verb_disp": 0}

        if opts is None:
            opts = {}

        for k, v in DEFAULTS.items():
            if k not in kwargs:
                opts[k] = v

        self.opts = opts

    def _setup(self, problem, **kwargs):
        xl, xu = problem.bounds()
        if self.normalize:
            self.norm, self.opts['bounds'] = bounds_if_normalize(xl, xu)
        else:
            self.norm = NoNormalization()
            self.opts['bounds'] = [xl, xu]
        self.opts['seed'] = self.seed

    def _initialize_advance(self, infills=None, **kwargs):
        super()._initialize_advance(infills, **kwargs)
        x = self.norm.forward(self.x0.X)
        self.es = cma.CMAEvolutionStrategy(x, self.sigma, inopts=self.opts)

    def _infill(self):
        X = self.norm.backward(np.array(self.es.ask()))
        return Population.new("X", X)

    def _advance(self, infills=None, **kwargs):
        X, F = infills.get("X", "F")
        X = self.norm.forward(X)

        self.es.tell(X, F[:, 0])
        self.pop = infills

        if self.es.stop():
            self.termination.force_termination = True

    def _set_optimum(self):
        pop = self.pop if self.opt is None else Population.merge(self.opt, self.pop)
        self.opt = filter_optimum(pop, least_infeasible=True)


class BIPOPCMAES(CMAES):

    def __init__(self, restarts=4, **kwargs):
        super().__init__(restarts=restarts, bipop=True, **kwargs)


def bounds_if_normalize(xl, xu):
    norm = ZeroToOneNormalization(xl=xl, xu=xu)

    _xl, _xu = np.zeros_like(xl), np.ones_like(xu)
    if xl is not None:
        _xl[np.isnan(xl)] = np.nan
    if xu is not None:
        _xu[np.isnan(xu)] = np.nan

    return norm, [_xl, _xu]


parse_doc_string(CMAES.__init__)



def install_matlab():
    print("Please install the Matlab python interface:")
    print("Tutorial: https://www.mathworks.com/help/matlab/matlab_external/install-the-matlab-engine-for-python.html")
    print("---------------------------")
    print("Go to:")
    print("Windows:", 'cd "matlabroot\extern\engines\python"')
    print("Linux/Mac:", 'cd "matlabroot/extern/engines/python"')
    print("python setup.py install")
    print("---------------------------")


class MatlabEngine:
    """
    Launching the Matlab engine can become time-consuming and thus shall only be done once and the instance
    could be reused by different kinds of problems at the same time.

    This is an implementation based on the singleton pattern where only one instance of the engine is used.

    """

    __instance = None

    @staticmethod
    def get_instance():
        if MatlabEngine.__instance is None:

            try:
                import matlab
                import matlab.engine
            except:
                print(install_matlab())

            MatlabEngine.__instance = matlab.engine.start_matlab(option="")

        return MatlabEngine.__instance

    @staticmethod
    def shutdown():
        MatlabEngine.__instance.quit()



import pymoo.gradient.toolbox as anp
import numpy as np

from pymoo.core.problem import Problem


class CantileveredBeam(Problem):

    def __init__(self):
        super().__init__(n_var=4, n_obj=1, n_ieq_constr=2, vtype=float)
        self.xl = np.array([2, 0.1, 0.1, 3.0])
        self.xu = np.array([12.0, 1.0, 2.0, 7.0])
        self.h1 = np.array([0.1, 0.25, 0.35, 0.5, 0.65, 0.75, 0.9, 1.0])

    def _evaluate(self, x, out, *args, **kwargs):
        E, L, P = 1e7, 36.0, 1000.0

        b1, h1, b2, H = x[:, 0], x[:, 1], x[:, 2], x[:, 3]
        I = 1 / 12 * b2 * (H - 2 * h1) ** 3 + 2 * (1 / 12 * b1 * h1 ** 3 + b1 * h1 * (H - h1) ** 2 / 4)
        volume = (2 * h1 * b1 + (H - 2 * h1) * b2) * L
        out["F"] = volume

        sigma = P * L * H / (2 * I)
        delta = P * L ** 3 / (3 * E * I)

        g1 = (sigma - 5000.0) / 5000.0
        g2 = (delta - 0.1) / 0.1
        out["G"] = anp.column_stack([g1, g2])

    def _calc_pareto_front(self):
        return 92.7693

    def _calc_pareto_set(self):
        return [9.4846, 0.1000, 0.1000, 7.0000]



from pymoo.indicators.gd import GD
from pymoo.indicators.hv import Hypervolume
from pymoo.indicators.igd import IGD
from pymoo.termination.ftol import MultiObjectiveSpaceTermination

from pymoo.util.display.column import Column
from pymoo.util.display.output import Output, pareto_front_if_possible
from pymoo.util.display.single import MinimumConstraintViolation, AverageConstraintViolation


class NumberOfNondominatedSolutions(Column):

    def __init__(self, width=6, **kwargs) -> None:
        super().__init__("n_nds", width=width, **kwargs)

    def update(self, algorithm):
        self.value = len(algorithm.opt)


class MultiObjectiveOutput(Output):

    def __init__(self):
        super().__init__()
        self.cv_min = MinimumConstraintViolation()
        self.cv_avg = AverageConstraintViolation()
        self.n_nds = NumberOfNondominatedSolutions()

        self.igd = Column("igd")
        self.gd = Column("gd")
        self.hv = Column("hv")
        self.eps = Column("eps")
        self.indicator = Column("indicator")

        self.pf = None
        self.indicator_no_pf = None

    def initialize(self, algorithm):
        problem = algorithm.problem

        self.columns += [self.n_nds]

        if problem.has_constraints():
            self.columns += [self.cv_min, self.cv_avg]

        self.pf = pareto_front_if_possible(problem)
        if self.pf is not None:
            self.columns += [self.igd, self.gd]

            if problem.n_obj == 2:
                self.columns += [self.hv]

        else:
            self.indicator_no_pf = MultiObjectiveSpaceTermination()
            self.columns += [self.eps, self.indicator]

    def update(self, algorithm):
        super().update(algorithm)

        for col in [self.igd, self.gd, self.hv, self.eps, self.indicator]:
            col.set(None)

        F, feas = algorithm.opt.get("F", "feas")
        F = F[feas]

        if len(F) > 0:

            if self.pf is not None:

                if feas.sum() > 0:
                    self.igd.set(IGD(self.pf, zero_to_one=True).do(F))
                    self.gd.set(GD(self.pf, zero_to_one=True).do(F))

                    if self.hv in self.columns:
                        self.hv.set(Hypervolume(pf=self.pf, zero_to_one=True).do(F))

            if self.indicator_no_pf is not None:

                ind = self.indicator_no_pf
                ind.update(algorithm)

                valid = ind.delta_ideal is not None

                if valid:

                    if ind.delta_ideal > ind.tol:
                        max_from = "ideal"
                        eps = ind.delta_ideal
                    elif ind.delta_nadir > ind.tol:
                        max_from = "nadir"
                        eps = ind.delta_nadir
                    else:
                        max_from = "f"
                        eps = ind.delta_f

                    self.eps.set(eps)
                    self.indicator.set(max_from)



import numpy as np

from pymoo.algorithms.base.genetic import GeneticAlgorithm
from pymoo.core.survival import Survival
from pymoo.docs import parse_doc_string
from pymoo.operators.crossover.sbx import SBX
from pymoo.operators.crossover.spx import SPX
from pymoo.operators.mutation.bitflip import BitflipMutation
from pymoo.operators.mutation.pm import PM
from pymoo.operators.sampling.rnd import BinaryRandomSampling
from pymoo.operators.sampling.rnd import FloatRandomSampling
from pymoo.operators.selection.tournament import compare, TournamentSelection
from pymoo.termination.default import DefaultSingleObjectiveTermination
from pymoo.util.display.single import SingleObjectiveOutput


# =========================================================================================================
# Survival
# =========================================================================================================

class FitnessSurvival(Survival):

    def __init__(self) -> None:
        super().__init__(filter_infeasible=False)

    def _do(self, problem, pop, n_survive=None, **kwargs):
        F, cv = pop.get("F", "cv")
        assert F.shape[1] == 1, "FitnessSurvival can only used for single objective single!"
        S = np.lexsort([F[:, 0], cv])
        pop.set("rank", np.argsort(S))
        return pop[S[:n_survive]]


# =========================================================================================================
# Implementation
# =========================================================================================================


def comp_by_cv_and_fitness(pop, P, **kwargs):
    S = np.full(P.shape[0], np.nan)

    for i in range(P.shape[0]):
        a, b = P[i, 0], P[i, 1]

        # if at least one solution is infeasible
        if pop[a].CV > 0.0 or pop[b].CV > 0.0:
            S[i] = compare(a, pop[a].CV, b, pop[b].CV, method='smaller_is_better', return_random_if_equal=True)

        # both solutions are feasible just set random
        else:
            S[i] = compare(a, pop[a].F, b, pop[b].F, method='smaller_is_better', return_random_if_equal=True)

    return S[:, None].astype(int)


class GA(GeneticAlgorithm):

    def __init__(self,
                 pop_size=100,
                 sampling=FloatRandomSampling(),
                 selection=TournamentSelection(func_comp=comp_by_cv_and_fitness),
                 crossover=SBX(),
                 mutation=PM(),
                 survival=FitnessSurvival(),
                 eliminate_duplicates=True,
                 n_offsprings=None,
                 output=SingleObjectiveOutput(),
                 **kwargs):
        super().__init__(pop_size=pop_size,
                         sampling=sampling,
                         selection=selection,
                         crossover=crossover,
                         mutation=mutation,
                         survival=survival,
                         eliminate_duplicates=eliminate_duplicates,
                         n_offsprings=n_offsprings,
                         output=output,
                         **kwargs)

        self.termination = DefaultSingleObjectiveTermination()


class BGA(GA):

    def __init__(self,
                 sampling=BinaryRandomSampling(),
                 crossover=SPX(),
                 mutation=BitflipMutation(),
                 **kwargs):
        super().__init__(sampling=sampling, crossover=crossover, mutation=mutation, **kwargs)


parse_doc_string(GA.__init__)



import numpy as np

import pymoo.gradient.toolbox as anp

from pymoo.core.meta import Meta
from pymoo.core.problem import Problem


class ConstraintsFromBounds(Meta, Problem):

    def __init__(self, problem, remove_bonds=False):
        super().__init__(problem)
        self.n_ieq_constr += 2 * self.n_var

        if remove_bonds:
            self.xl, self.xu = None, None

    def do(self, X, return_values_of, *args, **kwargs):

        out = self.__object__.do(X, return_values_of, *args, **kwargs)

        # get the boundaries for normalization
        xl, xu = self.bounds()

        # add the boundary constraint if enabled
        _G = anp.column_stack([xl - X, X - xu])

        out["G"] = anp.column_stack([out["G"], _G])

        if "dG" in out:
            _dG = np.zeros((len(X), 2 * self.n_var, self.n_var))
            _dG[:, :self.n_var, :] = - np.eye(self.n_var)
            _dG[:, self.n_var:, :] = np.eye(self.n_var)
            out["dG"] = np.column_stack([out["dG"], _dG])

        return out



import pymoo.gradient.toolbox as anp

from pymoo.core.problem import Problem


class MultiModalSimple1(Problem):
    def __init__(self, n_var=1):
        super().__init__(n_var=n_var, n_obj=1, xl=0, xu=1, vtype=float)

    def _evaluate(self, x, out, *args, **kwargs):
        out["F"] = 1 - anp.exp(-x ** 2) * anp.sin(2 * anp.pi * x) ** 2


class MultiModalSimple2(Problem):
    def __init__(self, n_var=1):
        super().__init__(n_var=n_var, n_obj=1, xl=-1, xu=0, vtype=float)

    def _evaluate(self, x, out, *args, **kwargs):
        x = - x
        out["F"] = 1.1 - anp.exp(-2 * x) * anp.sin(5 * anp.pi * x) ** 2



import time

from pymoo.core.termination import Termination
from pymoo.util.misc import time_to_int


class TimeBasedTermination(Termination):

    def __init__(self, max_time) -> None:
        super().__init__()
        if isinstance(max_time, str):
            self.max_time = time_to_int(max_time)
        elif isinstance(max_time, int) or isinstance(max_time, float):
            self.max_time = max_time
        else:
            raise Exception("Either provide the time as a string or an integer.")

    def _update(self, algorithm):
        elapsed = time.time() - algorithm.start_time
        return elapsed / self.max_time



import numpy as np

from pymoo.core.crossover import Crossover
from pymoo.core.variable import Real, get
from pymoo.operators.repair.bounds_repair import repair_random_init


def pcx(X, eta, zeta, index):
    eps = 1e-32

    # the number of parents to be considered
    n_parents, n_matings, n_var = X.shape

    # calculate the differences from all parents to index parent
    diff_to_index = X - X[index]
    dist_to_index = np.linalg.norm(diff_to_index, axis=-1)
    dist_to_index = np.maximum(eps, dist_to_index)

    # find the centroid of the parents
    centroid = np.mean(X, axis=0)

    # calculate the difference between the centroid and the k-th parent
    diff_to_centroid = centroid - X[index]

    dist_to_centroid = np.linalg.norm(diff_to_centroid, axis=-1)
    dist_to_centroid = np.maximum(eps, dist_to_centroid)

    # orthogonal directions are computed
    orth_dir = np.zeros_like(dist_to_index)

    for i in range(n_parents):
        if i != index:
            temp1 = (diff_to_index[i] * diff_to_centroid).sum(axis=-1)
            temp2 = temp1 / (dist_to_index[i] * dist_to_centroid)
            temp3 = np.maximum(0.0, 1.0 - temp2 ** 2)
            orth_dir[i] = dist_to_index[i] * (temp3 ** 0.5)

    # this is the avg of the perpendicular distances from other parents to the parent k
    D_not = orth_dir.sum(axis=0) / (n_parents - 1)

    # generating zero-mean normally distributed variables
    sigma = D_not[:, None] * eta.repeat(n_var, axis=1)
    rnd = np.random.normal(loc=0.0, scale=sigma)

    # implemented just like the c code - generate_new.h file
    inner_prod = np.sum(rnd * diff_to_centroid, axis=-1, keepdims=True)
    noise = rnd - (inner_prod * diff_to_centroid) / dist_to_centroid[:, None] ** 2

    bias_to_centroid = np.random.normal(0.0, zeta) * diff_to_centroid

    # the array which is finally returned
    Xp = X[index] + noise + bias_to_centroid

    return Xp


class ParentCentricCrossover(Crossover):
    def __init__(self,
                 eta=0.1,
                 zeta=0.1,
                 **kwargs):

        super().__init__(n_parents=3, n_offsprings=1, **kwargs)
        self.eta = Real(eta, bounds=(0.01, 0.3))
        self.zeta = Real(zeta, bounds=(0.01, 0.3))

    def _do(self, problem, X, params=None, **kwargs):
        n_parents, n_matings, n_var = X.shape
        zeta, eta = get(self.zeta, self.eta, size=(n_matings, 1))

        index = 0

        Xp = pcx(X, eta, zeta, index=index)

        if problem.has_bounds():
            Xp = repair_random_init(Xp, X[index], *problem.bounds())

        return Xp[None, :]


class PCX(ParentCentricCrossover):
    pass



import numpy as np

from pymoo.util.nds.non_dominated_sorting import NonDominatedSorting
from pymoo.problems.dyn import DynamicTestProblem
from pymoo.util.remote import Remote


class DF(DynamicTestProblem):

    def __init__(self, n_var=10, nt=10, taut=20, **kwargs):
        super().__init__(nt,
                         taut,
                         n_var=n_var,
                         n_obj=2,
                         xl=0,
                         xu=1,
                         **kwargs)

    def _calc_pareto_front(self, *args, **kwargs):
        return Remote.get_instance().load("pymoo", "pf", "DF", f"{str(self.__class__.__name__)}.pf")


class DF1(DF):

    def _evaluate(self, x, out, *args, **kwargs):
        v = np.sin(0.5 * np.pi * self.time)

        G = np.abs(v)
        H = 0.75 * v + 1.25
        g = 1 + np.sum((x[:, 1:] - G) ** 2, axis=1)

        f1 = x[:, 0]
        f2 = g * (1 - ((f1 / g) ** H))

        out["F"] = np.column_stack([f1, f2])

    def _calc_pareto_front(self, *args, n_pareto_points=100, **kwargs):
        v = np.sin(0.5 * np.pi * self.time)
        H = 0.75 * v + 1.25

        f1 = np.linspace(0, 1, n_pareto_points)
        return np.array([f1, 1 - f1 ** H]).T


class DF2(DF):
    def _evaluate(self, x, out, *args, **kwargs):
        v = np.sin(0.5 * np.pi * self.time)
        G = np.abs(v)

        n = self.n_var
        r = int((n - 1) * G)
        not_r = [k for k in range(n) if k != r]

        g = 1 + np.sum((x[:, not_r] - G) ** 2, axis=1)

        f1 = x[:, r]
        f2 = g * (1 - np.power(f1 / g, 0.5))

        out["F"] = np.column_stack([f1, f2])

    def _calc_pareto_front(self, *args, n_pareto_points=100, **kwargs):
        f1 = np.linspace(0, 1, n_pareto_points)
        return np.array([f1, 1 - np.sqrt(f1)]).T


class DF3(DF):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.xl[1:] = -1.0
        self.xu[1:] = 2.0

    def _evaluate(self, x, out, *args, **kwargs):
        G = np.sin(0.5 * np.pi * self.time)
        H = G + 1.5

        g = 1 + np.sum((x[:, 1:] - G - x[:, [0]] ** H) ** 2, axis=1)
        f1 = x[:, 0]
        f2 = g * (1 - (x[:, 0] / g) ** H)

        out["F"] = np.column_stack([f1, f2])

    def _calc_pareto_front(self, *args, n_pareto_points=100, **kwargs):
        x = np.linspace(0, 1, n_pareto_points)
        g = 1

        G = np.sin(np.dot(np.dot(0.5, np.pi), self.time))
        H = G + 1.5
        f1 = np.copy(x)
        f2 = np.dot(g, (1 - (x / g) ** H))

        return np.column_stack([f1, f2])


class DF4(DF):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.xl[:] = -2.0
        self.xu[:] = +2.0

    def _evaluate(self, x, out, *args, **kwargs):
        n = self.n_var
        a = np.sin(0.5 * np.pi * self.time)
        b = 1 + np.abs(np.cos(0.5 * np.pi * self.time))
        H = 1.5 + a
        c = np.maximum(np.abs(a), a + b)

        g = 1.0
        for i in range(1, n):
            g += (x[:, i] - (a * (x[:, 0] / c) ** 2 / (i + 1))) ** 2

        f1 = g * np.abs(x[:, 0] - a) ** H
        f2 = g * np.abs(x[:, 0] - a - b) ** H

        out["F"] = np.column_stack([f1, f2])

    def _calc_pareto_front(self, *args, n_pareto_points=100, **kwargs):
        a = np.sin(0.5 * np.pi * self.time)
        b = 1 + np.abs(np.cos(0.5 * np.pi * self.time))
        H = 1.5 + a
        x = np.linspace(a, a + b, n_pareto_points)

        f1 = np.abs(x - a) ** H
        f2 = np.abs(x - a - b) ** H

        return np.array([f1, f2]).T


class DF5(DF):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.xl[1:] = -1.0
        self.xu[1:] = +1.0

    def _evaluate(self, x, out, *args, **kwargs):
        G = np.sin(0.5 * np.pi * self.time)
        w = np.floor(10 * G)
        g = 1 + np.sum((x[:, 1:] - G) ** 2, axis=1)
        f1 = g * (x[:, 0] + 0.02 * np.sin(w * np.pi * x[:, 0]))
        f2 = g * (1 - x[:, 0] + 0.02 * np.sin(w * np.pi * x[:, 0]))

        out["F"] = np.column_stack([f1, f2])

    def _calc_pareto_front(self, *args, n_pareto_points=100, **kwargs):
        x = np.linspace(0, 1, n_pareto_points)
        G = np.sin(0.5 * np.pi * self.time)
        w = np.floor(10 * G)
        f1 = x + 0.02 * np.sin(w * np.pi * x)
        f2 = 1 - x + 0.02 * np.sin(w * np.pi * x)
        return np.array([f1, f2]).T


class DF6(DF):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.xl[1:] = -1.0
        self.xu[1:] = +1.0

    def _evaluate(self, x, out, *args, **kwargs):
        G = np.sin(0.5 * np.pi * self.time)
        a = 0.2 + 2.8 * np.abs(G)
        y = x[:, 1:] - G
        g = 1 + np.sum((np.abs(G) * y ** 2 - 10 * np.cos(2 * np.pi * y) + 10), axis=1)

        f1 = g * np.power(x[:, 0] + 0.1 * np.sin(3 * np.pi * x[:, 0]), a)
        f2 = g * np.power(1 - x[:, 0] + 0.1 * np.sin(3 * np.pi * x[:, 0]), a)

        out["F"] = np.column_stack([f1, f2])

    def _calc_pareto_front(self, *args, n_pareto_points=100, **kwargs):
        x = np.linspace(0, 1, n_pareto_points)
        G = np.sin(0.5 * np.pi * self.time)
        a = 0.2 + 2.8 * np.abs(G)
        f1 = (x + 0.1 * np.sin(3 * np.pi * x)) ** a
        f2 = (1 - x + 0.1 * np.sin(3 * np.pi * x)) ** a

        return np.array([f1, f2]).T


class DF7(DF):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.xl[0] = 1.0
        self.xu[0] = 4.0

    def _evaluate(self, x, out, *args, **kwargs):
        a = 5 * np.cos(0.5 * np.pi * self.time)
        g = 1 + np.sum((x[:, 1:] - 1 / (1 + np.exp(a * (x[:, [0]] - 2.5)))) ** 2, axis=1)

        f1 = g * (1 + self.time) / x[:, 0]
        f2 = g * x[:, 0] / (1 + self.time)

        out["F"] = np.column_stack([f1, f2])

    def _calc_pareto_front(self, *args, n_pareto_points=100, **kwargs):
        x = np.linspace(1, 4, n_pareto_points)
        f1 = (1 + self.time) / x
        f2 = x / (1 + self.time)
        pf = np.array([f1, f2]).T
        pf = pf[np.argsort(pf[:, 0])]
        return pf


class DF8(DF):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.xl[1:] = -1.0
        self.xu[1:] = +1.0

    def _evaluate(self, x, out, *args, **kwargs):
        G = np.sin(0.5 * np.pi * self.time)
        a = 2.25 + 2 * np.cos(2 * np.pi * self.time)
        b = 1
        tmp = G * np.sin(4 * np.pi * np.power(x[:, 0].reshape(len(x), 1), b)) / (1 + np.abs(G))
        g = 1 + np.sum((x[:, 1:] - tmp) ** 2, axis=1)
        f1 = g * (x[:, 0] + 0.1 * np.sin(3 * np.pi * x[:, 0]))
        f2 = g * np.power(1 - x[:, 0] + 0.1 * np.sin(3 * np.pi * x[:, 0]), a)

        out["F"] = np.column_stack([f1, f2])

    def _calc_pareto_front(self, *args, n_pareto_points=100, **kwargs):
        x = np.linspace(0, 1, n_pareto_points)
        a = 2.25 + 2 * np.cos(2 * np.pi * self.time)

        f1 = x + 0.1 * np.sin(3 * np.pi * x)
        f2 = (1 - x + 0.1 * np.sin(3 * np.pi * x)) ** a

        return np.array([f1, f2]).T


class DF9(DF):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.xl[1:] = -1.0
        self.xu[1:] = +1.0

    def _evaluate(self, x, out, *args, **kwargs):
        _, n = x.shape
        N = 1 + np.floor(10 * np.abs(np.sin(0.5 * np.pi * self.time)))
        g = 1
        for i in range(1, n):
            tmp = x[:, i] - np.cos(4 * self.time + x[:, 0] + x[:, i - 1])
            g = g + tmp ** 2
        f1 = g * (x[:, 0] + np.maximum(0, (0.1 + 0.5 / N) * np.sin(2 * N * np.pi * x[:, 0])))
        f2 = g * (1 - x[:, 0] + np.maximum(0, (0.1 + 0.5 / N) * np.sin(2 * N * np.pi * x[:, 0])))

        out["F"] = np.column_stack([f1, f2])

    def _calc_pareto_front(self, *args, n_pareto_points=100, **kwargs):
        x = np.linspace(0, 1, n_pareto_points)
        N = 1 + np.floor(10 * np.abs(np.sin(0.5 * np.pi * self.time)))

        f1 = x + np.maximum(0, (0.1 + 0.5 / N) * np.sin(2 * N * np.pi * x))
        f2 = 1 - x + np.maximum(0, (0.1 + 0.5 / N) * np.sin(2 * N * np.pi * x))

        h = get_PF(np.array([f1, f2]), True)
        return h


class DF10(DF):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.xl[2:] = -1.0
        self.xu[2:] = +1.0
        self.n_obj = 3

    def _evaluate(self, x, out, *args, **kwargs):
        G = np.sin(0.5 * np.pi * self.time)
        H = 2.25 + 2 * np.cos(0.5 * np.pi * self.time)
        x0 = x[:, 0].reshape(len(x), 1)
        x1 = x[:, 1].reshape(len(x), 1)
        tmp = np.sin(4 * np.pi * (x0 + x1)) / (1 + np.abs(G))  # in the document is 2*
        g = 1 + np.sum((x[:, 2:] - tmp) ** 2, axis=1)
        g = g.reshape(len(g), 1)
        f1 = (g * np.power(np.sin(0.5 * np.pi * x0), H)).reshape(len(g), )
        f2 = (g * np.power(np.sin(0.5 * np.pi * x1) * np.cos(0.5 * np.pi * x0), H)).reshape(len(g), )
        f3 = (g * np.power(np.cos(0.5 * np.pi * x1) * np.cos(0.5 * np.pi * x0), H)).reshape(len(g), )

        out["F"] = np.column_stack([f1, f2, f3])

    def _calc_pareto_front(self, *args, n_pareto_points=100, **kwargs):
        H = 20
        x1, x2 = np.meshgrid(np.linspace(0, 1, H), np.linspace(0, 1, H), indexing='xy')
        H = 2.25 + 2 * np.cos(0.5 * np.pi * self.time)
        g = 1
        f1 = g * np.sin(0.5 * np.pi * x1) ** H
        f2 = np.multiply(g * np.sin(0.5 * np.pi * x2) ** H, np.cos(0.5 * np.pi * x1) ** H)
        f3 = np.multiply(g * np.cos(0.5 * np.pi * x2) ** H, np.cos(0.5 * np.pi * x1) ** H)

        return get_PF(np.array([f1, f2, f3]), False)


class DF11(DF):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.n_obj = 3

    def _evaluate(self, x, out, *args, **kwargs):
        G = np.abs(np.sin(0.5 * np.pi * self.time))
        g = 1 + G + np.sum((x[:, 2:] - 0.5 * G * x[:, 0].reshape(len(x), 1)) ** 2, axis=1)
        y = [np.pi * G / 6.0 + (np.pi / 2 - np.pi * G / 3.0) * x[:, i] for i in [0, 1]]

        f1 = g * np.sin(y[0])
        f2 = g * np.sin(y[1]) * np.cos(y[0])
        f3 = g * np.cos(y[1]) * np.cos(y[0])

        out["F"] = np.column_stack([f1, f2, f3])

    def _calc_pareto_front(self, *args, n_pareto_points=100, **kwargs):
        H = 20
        x1, x2 = np.meshgrid(np.linspace(0, 1, H), np.linspace(0, 1, H), indexing='xy')
        G = np.abs(np.sin(0.5 * np.pi * self.time))
        y1 = np.pi * G / 6 + (np.pi / 2 - np.pi * G / 3) * x1
        y2 = np.pi * G / 6 + (np.pi / 2 - np.pi * G / 3) * x2

        f1 = np.sin(y1)
        f2 = np.dot(np.multiply(1, np.sin(y2)), np.cos(y1))
        f3 = np.dot(np.multiply(1, np.cos(y2)), np.cos(y1))

        return get_PF(np.array([f1, f2, f3]), False)


class DF12(DF):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.n_obj = 3
        self.xl[2:] = -1.0
        self.xu[2:] = +1.0

    def _evaluate(self, x, out, *args, **kwargs):
        k = 10 * np.sin(np.pi * self.time)
        #        r = 1 - np.modulo(k,2)
        r = 1
        x0 = x[:, 0].reshape(len(x), 1)
        tmp1 = x[:, 2:] - np.sin(self.time * x0)
        tmp2 = np.abs(np.sin(np.floor(k * (2 * x[:, 0:2] - r)) * np.pi / 2))
        g = 1 + np.sum(tmp1 ** 2, axis=1) + np.prod(tmp2)

        f1 = g * np.cos(0.5 * np.pi * x[:, 1]) * np.cos(0.5 * np.pi * x[:, 0])
        f2 = g * np.sin(0.5 * np.pi * x[:, 1]) * np.cos(0.5 * np.pi * x[:, 0])
        f3 = g * np.sin(0.5 * np.pi * x[:, 0])

        out["F"] = np.column_stack([f1, f2, f3])

    def _calc_pareto_front(self, *args, n_pareto_points=100, **kwargs):
        H = 20
        x1, x2 = np.meshgrid(np.linspace(0, 1, H), np.linspace(0, 1, H), indexing='xy')
        k = 10 * np.sin(np.pi * self.time)
        tmp2 = np.abs(
            (np.sin((np.floor(k * (2 * x1 - 1)) * np.pi) / 2) *
             np.sin((np.floor(k * (2 * x2 - 1)) * np.pi) / 2)))
        g = 1 + tmp2
        f1 = np.multiply(np.multiply(g, np.cos(0.5 * np.pi * x2)), np.cos(0.5 * np.pi * x1))
        f2 = np.multiply(np.multiply(g, np.sin(0.5 * np.pi * x2)), np.cos(0.5 * np.pi * x1))
        f3 = np.multiply(g, np.sin(0.5 * np.pi * x1))

        return get_PF(np.array([f1, f2, f3]), True)


class DF13(DF):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.n_obj = 3
        self.xl[2:] = -1.0
        self.xu[2:] = +1.0

    def _evaluate(self, x, out, *args, **kwargs):
        G = np.sin(0.5 * np.pi * self.time)
        p = np.floor(6 * G)
        x0 = x[:, 0].reshape(len(x), 1)
        x1 = x[:, 1].reshape(len(x), 1)
        g = 1 + np.sum((x[:, 2:] - G) ** 2, axis=1)
        g = g.reshape(len(g), 1)
        f1 = g * np.cos(0.5 * np.pi * x0) ** 2
        f2 = g * np.cos(0.5 * np.pi * x1) ** 2
        f3 = g * np.sin(0.5 * np.pi * x0) ** 2 + np.sin(0.5 * np.pi * x0) * np.cos(p * np.pi * x0) ** 2 + np.sin(
            0.5 * np.pi * x1) ** 2 + np.sin(0.5 * np.pi * x1) * np.cos(p * np.pi * x1) ** 2

        out["F"] = np.column_stack([f1, f2, f3])

    def _calc_pareto_front(self, *args, n_pareto_points=100, **kwargs):
        H = 20
        x1, x2 = np.meshgrid(np.linspace(0, 1, H), np.linspace(0, 1, H), indexing='xy')
        G = np.sin(0.5 * np.pi * self.time)
        p = np.floor(6 * G)

        f1 = np.cos(0.5 * np.pi * x1) ** 2
        f2 = np.cos(0.5 * np.pi * x2) ** 2
        f3 = np.sin(0.5 * np.pi * x1) ** 2 + np.sin(0.5 * np.pi * x1) * np.cos(p * np.pi * x1) ** 2 + np.sin(
            0.5 * np.pi * x2) ** 2 + \
             np.sin(0.5 * np.pi * x2) * np.cos(p * np.pi * x2) ** 2

        return get_PF(np.array([f1, f2, f3]), True)


class DF14(DF):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.n_obj = 3
        self.xl[2:] = -1.0
        self.xu[2:] = +1.0

    def _evaluate(self, x, out, *args, **kwargs):
        G = np.sin(0.5 * np.pi * self.time)
        x0 = x[:, 0].reshape(len(x), 1)
        x1 = x[:, 1].reshape(len(x), 1)

        g = 1 + np.sum((x[:, 2:] - G) ** 2, axis=1)
        g = g.reshape(len(g), 1)
        y = 0.5 + G * (x0 - 0.5)

        f1 = g * (1 - y + 0.05 * np.sin(6 * np.pi * y))
        f2 = g * (1 - x1 + 0.05 * np.sin(6 * np.pi * x1)) * (y + 0.05 * np.sin(6 * np.pi * y))
        f3 = g * (x1 + 0.05 * np.sin(6 * np.pi * x1)) * (y + 0.05 * np.sin(6 * np.pi * y))

        out["F"] = np.column_stack([f1, f2, f3])

    def _calc_pareto_front(self, *args, n_pareto_points=100, **kwargs):
        H = 20
        x1, x2 = np.meshgrid(np.linspace(0, 1, H), np.linspace(0, 1, H), indexing='xy')
        G = np.sin(0.5 * np.pi * self.time)
        y = 0.5 + G * (x1 - 0.5)
        f1 = 1 - y + 0.05 * np.sin(6 * np.pi * y)
        f2 = np.multiply(1 - x2 + 0.05 * np.sin(6 * np.pi * x2), y + 0.05 * np.sin(6 * np.pi * y))
        f3 = np.multiply(x2 + 0.05 * np.sin(6 * np.pi * x2), y + 0.05 * np.sin(6 * np.pi * y))

        return get_PF(np.array([f1, f2, f3]), False)


def get_PF(f=None, nondominate=None):
    nds = NonDominatedSorting()
    ncell = len(f)
    s = np.size(f[1])
    h = []
    for i in np.arange(ncell):
        fi = np.reshape(f[i], s, order='F')
        h.append(fi)
    h = np.array(h).T
    h = np.reshape(h, (s, ncell))

    if nondominate:
        fronts = nds.do(F=h, only_non_dominated_front=True)
        h = h[fronts]
    return h



import numpy as np

from pymoo.operators.repair.bounds_repair import BoundsRepair
from pymoo.util.misc import at_least_2d_array


def bounce_back(X, xl, xu):
    only_1d = (X.ndim == 1)
    X = at_least_2d_array(X)

    xl = np.repeat(xl[None, :], X.shape[0], axis=0)
    xu = np.repeat(xu[None, :], X.shape[0], axis=0)

    # otherwise bounds back into the feasible space
    _range = xu - xl
    X[X < xl] = (xl + np.mod((xl - X), _range))[X < xl]
    X[X > xu] = (xu - np.mod((X - xu), _range))[X > xu]

    if only_1d:
        return X[0, :]
    else:
        return X


def bounce_back_by_problem(problem, X):
    return bounce_back(X, problem.xl, problem.xu)


class BounceBackOutOfBoundsRepair(BoundsRepair):

    def repair_out_of_bounds(self, problem, X, **kwargs):
        return bounce_back_by_problem(problem, X)



from pymoo.core.termination import Termination
from pymoo.termination.cv import ConstraintViolationTermination
from pymoo.termination.ftol import SingleObjectiveSpaceTermination, MultiObjectiveSpaceTermination
from pymoo.termination.max_eval import MaximumFunctionCallTermination
from pymoo.termination.max_gen import MaximumGenerationTermination
from pymoo.termination.robust import RobustTermination
from pymoo.termination.xtol import DesignSpaceTermination


class DefaultTermination(Termination):

    def __init__(self, x, cv, f, n_max_gen=1000, n_max_evals=100000) -> None:
        super().__init__()
        self.x = x
        self.cv = cv
        self.f = f

        self.max_gen = MaximumGenerationTermination(n_max_gen)
        self.max_evals = MaximumFunctionCallTermination(n_max_evals)

        self.criteria = [self.x, self.cv, self.f, self.max_gen, self.max_evals]

    def _update(self, algorithm):
        p = [criterion.update(algorithm) for criterion in self.criteria]
        return max(p)


class DefaultSingleObjectiveTermination(DefaultTermination):

    def __init__(self, xtol=1e-8, cvtol=1e-8, ftol=1e-6, period=30, **kwargs) -> None:
        x = RobustTermination(DesignSpaceTermination(xtol), period=period)
        cv = RobustTermination(ConstraintViolationTermination(cvtol, terminate_when_feasible=False), period=period)
        f = RobustTermination(SingleObjectiveSpaceTermination(ftol, only_feas=True), period=period)
        super().__init__(x, cv, f, **kwargs)


class DefaultMultiObjectiveTermination(DefaultTermination):

    def __init__(self, xtol=0.0005, cvtol=1e-8, ftol=0.005, n_skip=5, period=50, **kwargs) -> None:
        x = RobustTermination(DesignSpaceTermination(tol=xtol, n_skip=n_skip), period)
        cv = RobustTermination(ConstraintViolationTermination(cvtol, terminate_when_feasible=False, n_skip=n_skip), period)
        f = RobustTermination(MultiObjectiveSpaceTermination(ftol, only_feas=True, n_skip=n_skip), period)
        super().__init__(x, cv, f, **kwargs)





from pymoo.algorithms.base.line import LineSearch
from pymoo.core.evaluator import Evaluator
from pymoo.core.solution import Solution, SolutionSet
from pymoo.core.termination import NoTermination
from pymoo.optimize import minimize
from pymoo.problems.single import Sphere


class BacktrackingLineSearch(LineSearch):

    def __init__(self, alpha=0.1, beta=0.33, **kwargs):
        super().__init__(**kwargs)
        self.alpha = alpha
        self.beta = beta
        self.termination = NoTermination()

    def _initialize_infill(self):
        super()._initialize_infill()
        self.t = 1.0
        self.point.set("t", self.t)

        # TODO:  evaluate_values_of needs to be moved into evaluator to be more flexible
        if self.point.get("dF") is None:
            self.evaluator.eval(self.problem, self.point, evaluate_values_of=["dF"], algorithm=self)

    def step(self):
        t = self.t
        x, f, df, p = self.point.X, self.point.F[0], self.point.get("dF")[0], self.direction

        infill = Solution(X=x + t * p, t=t)
        self.evaluator.eval(self.problem, infill, algorithm=self, evaluate_values_of=["F"])
        _f = infill.F[0]

        self.pop = SolutionSet.merge(self.pop, infill)
        self.infill = infill

        if _f < f + self.alpha * self.t * df.T @ p or self.t <= 1e-8:
            self.termination.force_termination = True
        else:
            self.t = self.t * self.beta


if __name__ == '__main__':
    import numpy as np

    problem = Sphere()

    X = np.array(np.random.random(problem.n_var))

    point = Solution(X=X)
    Evaluator(evaluate_values_of=["F", "dF"]).eval(problem, point)

    direction = - point.get("dF")[0]

    algorithm = BacktrackingLineSearch().setup(problem, point=point, direction=direction)._initialize()

    res = minimize(problem, algorithm)

    print(res.X)



import numpy as np


def calc_distance_to_weights(F, weights, utopian_point=None):
    norm = np.linalg.norm(weights, axis=1)

    if utopian_point is not None:
        F = F - utopian_point

    d1 = (F * weights).sum(axis=1) / norm
    d2 = np.linalg.norm(F - (d1[:, None] * weights / norm[:, None]), axis=1)

    return d1, d2



from pymoo.core.termination import Termination


class MaximumFunctionCallTermination(Termination):

    def __init__(self, n_max_evals=float("inf")) -> None:
        super().__init__()
        self.n_max_evals = n_max_evals

    def _update(self, algorithm):
        if self.n_max_evals is None:
            return 0.0
        else:
            return algorithm.evaluator.n_eval / self.n_max_evals



import pymoo.gradient.toolbox as anp
import numpy as np

from pymoo.core.problem import Problem


class Schwefel(Problem):
    def __init__(self, n_var=2):
        super().__init__(n_var=n_var, n_obj=1, n_ieq_constr=0, xl=-500, xu=500, vtype=float)

    def _evaluate(self, x, out, *args, **kwargs):
        out["F"] = 418.9829 * self.n_var - anp.sum(x * anp.sin(anp.sqrt(anp.abs(x))), axis=1)

    def _calc_pareto_front(self):
        return 0

    def _calc_pareto_set(self):
        return np.full(self.n_var, 420.9687)



import numpy as np

from pymoo.core.individual import Individual
from pymoo.core.population import Population
from pymoo.util.nds.non_dominated_sorting import NonDominatedSorting


def filter_optimum(pop, least_infeasible=False):

    # if the population is none to optimum can be found
    if pop is None or len(pop) == 0:
        return None

    # first only choose feasible solutions
    ret = pop[pop.get("feas")]

    # if at least one feasible solution was found
    if len(ret) > 0:

        # then check the objective values
        F = ret.get("F")

        if F.shape[1] > 1:
            I = NonDominatedSorting().do(F, only_non_dominated_front=True)
            ret = ret[I]

        else:
            ret = ret[np.argmin(F[:, 0])]

    # no feasible solution was found
    else:
        # if flag enable report the least infeasible
        if least_infeasible:
            ret = pop[np.argmin(pop.get("CV"))]
        # otherwise just return none
        else:
            ret = None

    if isinstance(ret, Individual):
        ret = Population().create(ret)

    return ret


from math import sqrt, log, exp

import numpy as np

from pymoo.algorithms.soo.nonconvex.es import es_sigma, es_mut_repair
from pymoo.algorithms.soo.nonconvex.sres import SRES
from pymoo.core.population import Population
from pymoo.docs import parse_doc_string


class ISRES(SRES):

    def __init__(self, gamma=0.85, alpha=0.2, **kwargs):
        """
        Improved Stochastic Ranking Evolutionary Strategy (SRES)

        Parameters
        ----------
        alpha : float
            Length scale of the differentials during mutation.
        PF: float
            The stochastic ranking weight for choosing a random decision while doing the modified bubble sort.
        """

        super().__init__(**kwargs)
        self.gamma = gamma
        self.alpha = alpha

    def _setup(self, problem, **kwargs):
        super()._setup(problem, **kwargs)
        n = problem.n_var

        chi = (1 / (2 * n) + 1 / (2 * (n ** 0.5)))
        varphi = sqrt((2 / chi) * log((1 / self.alpha) * (exp(self.phi ** 2 * chi / 2) - (1 - self.alpha))))

        self.taup = varphi / ((2 * n) ** 0.5)
        self.tau = varphi / ((2 * (n ** 0.5)) ** 0.5)

    def _infill(self):
        pop, mu, _lambda = self.pop, self.pop_size, self.n_offsprings
        xl, xu = self.problem.bounds()
        X, sigma = pop.get("X", "sigma")

        # cycle through the elites individuals for create the solutions
        I = np.arange(_lambda) % min(mu, len(X))

        # transform X and sigma to the shape of number of offsprings
        X, sigma = X[I], sigma[I]

        # copy the original sigma to sigma prime to be modified
        Xp, sigmap = np.copy(X), np.copy(sigma)

        # for the best individuals do differential variation to provide a direction to search in
        Xp[:mu - 1] = X[:mu - 1] + self.gamma * (X[0] - X[1:mu])

        # update the sigma values for elite and non-elite individuals
        sigmap[mu - 1:] = np.minimum(self.sigma_max, es_sigma(sigma[mu - 1:], self.tau, self.taup))

        # execute the evolutionary strategy to calculate the offspring solutions
        Xp[mu - 1:] = X[mu - 1:] + sigmap[mu - 1:] * np.random.normal(size=sigmap[mu - 1:].shape)

        # repair the individuals which are not feasible by sampling from sigma again
        Xp = es_mut_repair(Xp, X, sigmap, xl, xu, 10)

        # now update the sigma values of the non-elites only
        sigmap[mu:] = sigma[mu:] + self.alpha * (sigmap[mu:] - sigma[mu:])

        # create the population to proceed further
        off = Population.new(X=Xp, sigma=sigmap)

        return off


parse_doc_string(ISRES.__init__)



import pymoo.gradient.toolbox as anp

from pymoo.core.problem import Problem
from pymoo.util.remote import Remote


class Kursawe(Problem):
    def __init__(self):
        super().__init__(n_var=3, n_obj=2, xl=-5, xu=5, vtype=float)

    def _evaluate(self, x, out, *args, **kwargs):
        l = []
        for i in range(2):
            l.append(-10 * anp.exp(-0.2 * anp.sqrt(anp.square(x[:, i]) + anp.square(x[:, i + 1]))))
        f1 = anp.sum(anp.column_stack(l), axis=1)

        f2 = anp.sum(anp.power(anp.abs(x), 0.8) + 5 * anp.sin(anp.power(x, 3)), axis=1)

        out["F"] = anp.column_stack([f1, f2])

    def _calc_pareto_front(self, *args, **kwargs):
        return Remote.get_instance().load("pymoo", "pf", "kursawe.pf")






import time

import numpy as np
from cma import CMAOptions
from cma import optimization_tools as ot
from cma.evolution_strategy import cma_default_options, CMAEvolutionStrategy
from cma.utilities import utils
from cma.utilities.math import Mh

all_stoppings = []

def void(_):
    pass

def my_fmin(x0,
            sigma0,
            objective_function=void,
            options=None,
            args=(),
            gradf=None,
            restarts=0,
            restart_from_best='False',
            incpopsize=2,
            eval_initial_x=False,
            parallel_objective=None,
            noise_handler=None,
            noise_change_sigma_exponent=1,
            noise_kappa_exponent=0,  # TODO: add max kappa value as parameter
            bipop=False,
            callback=None):

    if 1 < 3:  # try: # pass on KeyboardInterrupt
        if not objective_function and not parallel_objective:  # cma.fmin(0, 0, 0)
            return CMAOptions()  # these opts are by definition valid

        fmin_options = locals().copy()  # archive original options
        del fmin_options['objective_function']
        del fmin_options['x0']
        del fmin_options['sigma0']
        del fmin_options['options']
        del fmin_options['args']

        if options is None:
            options = cma_default_options
        CMAOptions().check_attributes(options)  # might modify options
        # checked that no options.ftarget =
        opts = CMAOptions(options.copy()).complement()

        if callback is None:
            callback = []
        elif callable(callback):
            callback = [callback]

        # BIPOP-related variables:
        runs_with_small = 0
        small_i = []
        large_i = []
        popsize0 = None  # to be evaluated after the first iteration
        maxiter0 = None  # to be evaluated after the first iteration
        base_evals = 0

        irun = 0
        best = ot.BestSolution()
        all_stoppings = []
        while True:  # restart loop
            sigma_factor = 1

            # Adjust the population according to BIPOP after a restart.
            if not bipop:
                # BIPOP not in use, simply double the previous population
                # on restart.
                if irun > 0:
                    popsize_multiplier = fmin_options['incpopsize'] ** (irun - runs_with_small)
                    opts['popsize'] = popsize0 * popsize_multiplier

            elif irun == 0:
                # Initial run is with "normal" population size; it is
                # the large population before first doubling, but its
                # budget accounting is the same as in case of small
                # population.
                poptype = 'small'

            elif sum(small_i) < sum(large_i):
                # An interweaved run with small population size
                poptype = 'small'
                if 11 < 3:  # not needed when compared to irun - runs_with_small
                    restarts += 1  # A small restart doesn't count in the total
                runs_with_small += 1  # _Before_ it's used in popsize_lastlarge

                sigma_factor = 0.01 ** np.random.uniform()  # Local search
                popsize_multiplier = fmin_options['incpopsize'] ** (irun - runs_with_small)
                opts['popsize'] = np.floor(popsize0 * popsize_multiplier ** (np.random.uniform() ** 2))
                opts['maxiter'] = min(maxiter0, 0.5 * sum(large_i) / opts['popsize'])
                # print('small basemul %s --> %s; maxiter %s' % (popsize_multiplier, opts['popsize'], opts['maxiter']))

            else:
                # A run with large population size; the population
                # doubling is implicit with incpopsize.
                poptype = 'large'

                popsize_multiplier = fmin_options['incpopsize'] ** (irun - runs_with_small)
                opts['popsize'] = popsize0 * popsize_multiplier
                opts['maxiter'] = maxiter0
                # print('large basemul %s --> %s; maxiter %s' % (popsize_multiplier, opts['popsize'], opts['maxiter']))

            if not callable(objective_function) and callable(parallel_objective):
                def objective_function(x, *args):
                    """created from `parallel_objective` argument"""
                    return parallel_objective([x], *args)[0]

            # recover from a CMA object
            if irun == 0 and isinstance(x0, MyCMAEvolutionStrategy):
                es = x0
                x0 = es.inputargs['x0']  # for the next restarts
                if np.isscalar(sigma0) and np.isfinite(sigma0) and sigma0 > 0:
                    es.sigma = sigma0
                # debatable whether this makes sense:
                sigma0 = es.inputargs['sigma0']  # for the next restarts
                if options is not None:
                    es.opts.set(options)
                # ignore further input args and keep original options
            else:  # default case
                if irun and eval(str(fmin_options['restart_from_best'])):
                    utils.print_warning('CAVE: restart_from_best is often not useful',
                                        verbose=opts['verbose'])
                    es = MyCMAEvolutionStrategy(best.x, sigma_factor * sigma0, opts)
                else:
                    es = MyCMAEvolutionStrategy(x0, sigma_factor * sigma0, opts)
                # return opts, es
                if callable(objective_function) and (
                        eval_initial_x
                        or es.opts['CMA_elitist'] == 'initial'
                        or (es.opts['CMA_elitist'] and
                            eval_initial_x is None)):
                    x = es.gp.pheno(es.mean,
                                    into_bounds=es.boundary_handler.repair,
                                    archive=es.sent_solutions)
                    es.f0 = yield x
                    es.best.update([x], es.sent_solutions,
                                   [es.f0], 1)
                    es.countevals += 1
            es.objective_function = objective_function  # only for the record

            opts = es.opts  # processed options, unambiguous
            # a hack:
            fmin_opts = CMAOptions("unchecked", **fmin_options.copy())
            for k in fmin_opts:
                # locals() cannot be modified directly, exec won't work
                # in 3.x, therefore
                fmin_opts.eval(k, loc={'N': es.N,
                                       'popsize': opts['popsize']},
                               correct_key=False)

            es.logger.append = opts['verb_append'] or es.countiter > 0 or irun > 0
            # es.logger is "the same" logger, because the "identity"
            # is only determined by the `verb_filenameprefix` option
            logger = es.logger  # shortcut
            try:
                logger.persistent_communication_dict.update(
                    {'variable_annotations':
                         objective_function.variable_annotations})
            except AttributeError:
                pass

            if 11 < 3:
                if es.countiter == 0 and es.opts['verb_log'] > 0 and \
                        not es.opts['verb_append']:
                    logger = CMADataLogger(es.opts['verb_filenameprefix']
                                           ).register(es)
                    logger.add()
                es.writeOutput()  # initial values for sigma etc

            if noise_handler:
                if isinstance(noise_handler, type):
                    noisehandler = noise_handler(es.N)
                else:
                    noisehandler = noise_handler
                noise_handling = True
                if fmin_opts['noise_change_sigma_exponent'] > 0:
                    es.opts['tolfacupx'] = inf
            else:
                noisehandler = ot.NoiseHandler(es.N, 0)  # switched off
                noise_handling = False
            es.noise_handler = noisehandler

            # the problem: this assumes that good solutions cannot take longer than bad ones:
            # with EvalInParallel(objective_function, 2, is_feasible=opts['is_feasible']) as eval_in_parallel:
            if 1 < 3:
                while not es.stop():  # iteration loop
                    # X, fit = eval_in_parallel(lambda: es.ask(1)[0], es.popsize, args, repetitions=noisehandler.evaluations-1)
                    X, fit = yield from es.ask_and_eval(parallel_objective or objective_function,
                                             args, gradf=gradf,
                                             evaluations=noisehandler.evaluations,
                                             aggregation=np.median,
                                             parallel_mode=parallel_objective)  # treats NaN with resampling if not parallel_mode
                    # TODO: check args and in case use args=(noisehandler.evaluations, )

                    if 11 < 3 and opts['vv']:  # inject a solution
                        # use option check_point = [0]
                        if 0 * np.random.randn() >= 0:
                            X[0] = 0 + opts['vv'] * es.sigma ** 0 * np.random.randn(es.N)
                            fit[0] = yield X[0]
                            # print fit[0]
                    if es.opts['verbose'] > 4:  # may be undesirable with dynamic fitness (e.g. Augmented Lagrangian)
                        if es.countiter < 2 or min(fit) <= es.best.last.f:
                            degrading_iterations_count = 0  # comes first to avoid code check complaint
                        else:  # min(fit) > es.best.last.f:
                            degrading_iterations_count += 1
                            if degrading_iterations_count > 4:
                                utils.print_message('%d f-degrading iterations (set verbose<=4 to suppress)'
                                                    % degrading_iterations_count,
                                                    iteration=es.countiter)
                    es.tell(X, fit)  # prepare for next iteration
                    if noise_handling:  # it would be better to also use these f-evaluations in tell
                        es.sigma *= noisehandler(X, fit, objective_function, es.ask,
                                                 args=args) ** fmin_opts['noise_change_sigma_exponent']

                        es.countevals += noisehandler.evaluations_just_done  # TODO: this is a hack, not important though
                        # es.more_to_write.append(noisehandler.evaluations_just_done)
                        if noisehandler.maxevals > noisehandler.minevals:
                            es.more_to_write.append(noisehandler.evaluations)
                        if 1 < 3:
                            # If sigma was above multiplied by the same
                            #  factor cmean is divided by here, this is
                            #  like only multiplying kappa instead of
                            #  changing cmean and sigma.
                            es.sp.cmean *= np.exp(-noise_kappa_exponent * np.tanh(noisehandler.noiseS))
                            es.sp.cmean[es.sp.cmean > 1] = 1.0  # also works with "scalar arrays" like np.array(1.2)
                    for f in callback:
                        f is None or f(es)
                    es.disp()
                    logger.add(
                        # more_data=[noisehandler.evaluations, 10**noisehandler.noiseS] if noise_handling else [],
                        modulo=1 if es.stop() and logger.modulo else None)
                    if (opts['verb_log'] and opts['verb_plot'] and
                            (es.countiter % max(opts['verb_plot'], opts['verb_log']) == 0 or es.stop())):
                        logger.plot(324)

            # end while not es.stop
            if opts['eval_final_mean'] and callable(objective_function):
                mean_pheno = es.gp.pheno(es.mean,
                                         into_bounds=es.boundary_handler.repair,
                                         archive=es.sent_solutions)
                fmean = yield mean_pheno
                es.countevals += 1
                es.best.update([mean_pheno], es.sent_solutions, [fmean], es.countevals)

            best.update(es.best, es.sent_solutions)  # in restarted case
            # es.best.update(best)

            this_evals = es.countevals - base_evals
            base_evals = es.countevals

            # BIPOP stats update

            if irun == 0:
                popsize0 = opts['popsize']
                maxiter0 = opts['maxiter']
                # XXX: This might be a bug? Reproduced from Matlab
                # small_i.append(this_evals)

            if bipop:
                if poptype == 'small':
                    small_i.append(this_evals)
                else:  # poptype == 'large'
                    large_i.append(this_evals)

            # final message
            if opts['verb_disp']:
                es.result_pretty(irun, time.asctime(time.localtime()),
                                 best.f)

            irun += 1
            # if irun > fmin_opts['restarts'] or 'ftarget' in es.stop() \
            # if irun > restarts or 'ftarget' in es.stop() \
            all_stoppings.append(dict(es.stop(check=False)))  # keeping the order
            if irun - runs_with_small > fmin_opts['restarts'] or 'ftarget' in es.stop() \
                    or 'maxfevals' in es.stop(check=False) or 'callback' in es.stop(check=False):
                break
            opts['verb_append'] = es.countevals
            opts['popsize'] = fmin_opts['incpopsize'] * es.sp.popsize  # TODO: use rather options?
            try:
                opts['seed'] += 1
            except TypeError:
                pass

        # while irun

        # es.out['best'] = best  # TODO: this is a rather suboptimal type for inspection in the shell
        if irun:
            es.best.update(best)
            # TODO: there should be a better way to communicate the overall best
        return es.result + (es.stop(), es, logger)
        ### 4560
        # TODO refine output, can #args be flexible?
        # is this well usable as it is now?
    else:  # except KeyboardInterrupt:  # Exception as e:
        if eval(safe_str(options['verb_disp'])) > 0:
            print(' in/outcomment ``raise`` in last line of cma.fmin to prevent/restore KeyboardInterrupt exception')
        raise KeyboardInterrupt  # cave: swallowing this exception can silently mess up experiments, if ctrl-C is hit


class MyCMAEvolutionStrategy(CMAEvolutionStrategy):

    def ask_and_eval(self, func, args=(), gradf=None, number=None, xmean=None, sigma_fac=1,
                     evaluations=1, aggregation=np.median, kappa=1, parallel_mode=False):

        # initialize
        popsize = self.sp.popsize
        if number is not None:
            popsize = int(number)

        if self.opts['CMA_mirrormethod'] == 1:  # direct selective mirrors
            nmirrors = Mh.sround(self.sp.lam_mirr * popsize / self.sp.popsize)
            self._mirrormethod1_done = self.countiter
        else:
            # method==0 unconditional mirrors are done in ask_geno
            # method==2 delayed selective mirrors are done via injection
            nmirrors = 0
        assert nmirrors <= popsize // 2
        self.mirrors_idx = np.arange(nmirrors)  # might never be used
        is_feasible = self.opts['is_feasible']

        # do the work
        fit = []  # or np.NaN * np.empty(number)
        X_first = self.ask(popsize, xmean=xmean, gradf=gradf, args=args)
        if xmean is None:
            xmean = self.mean  # might have changed in self.ask
        X = []
        if parallel_mode:
            if hasattr(func, 'evaluations'):
                evals0 = func.evaluations
            fit_first = yield X_first
            # the rest is only book keeping and warnings spitting
            if hasattr(func, 'evaluations'):
                self.countevals += func.evaluations - evals0 - self.popsize  # why not .sp.popsize ?
            if nmirrors and self.opts['CMA_mirrormethod'] > 0 and self.countiter < 2:
                utils.print_warning(
                    "selective mirrors will not work in parallel mode",
                    "ask_and_eval", "CMAEvolutionStrategy")
            if evaluations > 1 and self.countiter < 2:
                utils.print_warning(
                    "aggregating evaluations will not work in parallel mode",
                    "ask_and_eval", "CMAEvolutionStrategy")
        else:
            fit_first = len(X_first) * [None]
        for k in range(popsize):
            x, f = X_first.pop(0), fit_first.pop(0)
            rejected = -1
            while f is None or not is_feasible(x, f):  # rejection sampling
                if parallel_mode:
                    utils.print_warning(
                        "rejection sampling will not work in parallel mode"
                        " unless the parallel_objective makes a distinction\n"
                        "between called with a numpy array vs a list (of"
                        " numpy arrays) as first argument.",
                        "ask_and_eval", "CMAEvolutionStrategy")
                rejected += 1
                if rejected:  # resample
                    x = self.ask(1, xmean, sigma_fac)[0]
                elif k >= popsize - nmirrors:  # selective mirrors
                    if k == popsize - nmirrors:
                        self.mirrors_idx = np.argsort(fit)[-1:-1 - nmirrors:-1]
                    x = self.get_mirror(X[self.mirrors_idx[popsize - 1 - k]])

                # constraints handling test hardwired ccccccccccc

                length_normalizer = 1
                # zzzzzzzzzzzzzzzzzzzzzzzzz
                if 11 < 3:
                    # for some unclear reason, this normalization does not work as expected: the step-size
                    # becomes sometimes too large and overall the mean might diverge. Is the reason that
                    # we observe random fluctuations, because the length is not selection relevant?
                    # However sigma-adaptation should mainly work on the correlation, not the length?
                    # Or is the reason the deviation of the direction introduced by using the original
                    # length, which also can effect the measured correlation?
                    # Update: if the length of z in CSA is clipped at chiN+1, it works, but only sometimes?
                    length_normalizer = self.N**0.5 / self.mahalanobis_norm(x - xmean)  # self.const.chiN < N**0.5, the constant here is irrelevant (absorbed by kappa)
                    # print(self.N**0.5 / self.mahalanobis_norm(x - xmean))
                    # self.more_to_write += [length_normalizer * 1e-3, length_normalizer * self.mahalanobis_norm(x - xmean) * 1e2]
                    if kappa == 1:
                        f = yield x
                    else:
                        f = yield xmean + kappa * length_normalizer * (x - xmean)

                if is_feasible(x, f) and evaluations > 1:
                    if kappa == 1:
                        _f = yield x
                        f = aggregation([f] + [_f])
                    else:
                        _f = []
                        for _i in range(int(evaluations - 1)):
                            v = yield xmean + kappa * length_normalizer * (x - xmean)
                            _f.append(v)
                        f = aggregation([f] + _f)

                if (rejected + 1) % 1000 == 0:
                    utils.print_warning('  %d solutions rejected (f-value NaN or None) at iteration %d' %
                          (rejected, self.countiter))
            fit.append(f)
            X.append(x)
        self.evaluations_per_f_value = int(evaluations)
        if any(f is None or utils.is_nan(f) for f in fit):
            idxs = [i for i in range(len(fit))
                    if fit[i] is None or utils.is_nan(fit[i])]
            utils.print_warning("f-values %s contain None or NaN at indices %s"
                                % (str(fit[:30]) + ('...' if len(fit) > 30 else ''),
                                   str(idxs)),
                                'ask_and_tell',
                                'CMAEvolutionStrategy',
                                self.countiter)
        return X, fit


import numpy as np
from scipy import special


class DasDennis:

    def __init__(self, n_partitions, n_dim, scaling=None):
        super().__init__()
        self.n_partitions = n_partitions
        self.n_dim = n_dim
        self.scaling = scaling

        self.stack = []
        self.stack.append(([], self.n_partitions))

    def number_of_points(self):
        return int(special.binom(self.n_dim + self.n_partitions - 1, self.n_partitions))

    def next(self, n_points=None):
        ret = []
        self.traverse(lambda p: ret.append(p), n_points)
        return np.array(ret)

    def has_next(self):
        return len(self.stack) > 0

    def traverse(self, func, n_points=None):

        if self.n_partitions == 0:
            return np.full((1, self.n_dim), 1 / self.n_dim)

        counter = 0

        while (n_points is None or counter < n_points) and len(self.stack) > 0:

            point, beta = self.stack.pop()

            if len(point) + 1 == self.n_dim:
                point.append(beta / (1.0 * self.n_partitions))

                if self.scaling is not None:
                    point = [p * self.scaling + ((1 - self.scaling) / len(point)) for p in point]

                func(point)
                counter += 1
            else:
                for i in range(beta + 1):
                    _point = list(point)
                    _point.append(1.0 * i / (1.0 * self.n_partitions))
                    self.stack.append((_point, beta - i))

        return counter



import numpy as np

from pymoo.indicators.hv import Hypervolume
from pymoo.indicators.igd import IGD
from pymoo.util.normalization import normalize
from pymoo.termination.delta import DeltaToleranceTermination


def calc_delta(a, b):
    return np.max(np.abs((a - b)))


def calc_delta_norm(a, b, norm):
    return np.max(np.abs((a - b) / norm))


class SingleObjectiveSpaceTermination(DeltaToleranceTermination):

    def __init__(self, tol=1e-6, only_feas=True, **kwargs) -> None:
        super().__init__(tol, **kwargs)
        self.only_feas = only_feas

    def _delta(self, prev, current):
        if prev == np.inf or current == np.inf:
            return np.inf
        else:
            return max(0, prev - current)

    def _data(self, algorithm):
        opt = algorithm.opt
        f = opt.get("f")

        if self.only_feas:
            f = f[opt.get("feas")]

        if len(f) > 0:
            return f.min()
        else:
            return np.inf


class MultiObjectiveSpaceTermination(DeltaToleranceTermination):

    def __init__(self, tol=0.0025, only_feas=True, **kwargs):
        super().__init__(tol, **kwargs)
        self.delta_ideal = None
        self.delta_nadir = None
        self.delta_f = None
        self.only_feas = only_feas

    def _data(self, algorithm):
        feas, F = algorithm.opt.get("feas", "F")

        if self.only_feas:
            F = F[feas]

        if len(F) > 0:
            return dict(ideal=F.min(axis=0), nadir=F.max(axis=0), F=F, feas=True)
        else:
            return dict(ideal=None, nadir=None, F=F, feas=False)

    def _delta(self, prev, current):

        if not (prev["feas"] and current["feas"]):
            return np.inf

        # this is the range between the nadir and the ideal point
        norm = current["nadir"] - current["ideal"]

        # if the range is degenerated (very close to zero) - disable normalization by dividing by one
        norm[norm < 1e-32] = 1.0

        # calculate the change from last to current in ideal and nadir point
        delta_ideal = calc_delta_norm(current["ideal"], prev["ideal"], norm)
        delta_nadir = calc_delta_norm(current["nadir"], prev["nadir"], norm)

        # get necessary data from the current population
        c_F, c_ideal, c_nadir = current["F"], current["ideal"], current["nadir"]
        p_F = prev["F"]

        # normalize last and current with respect to most recent ideal and nadir
        c_N = normalize(c_F, c_ideal, c_nadir)
        p_N = normalize(p_F, c_ideal, c_nadir)

        # calculate IGD from one to another
        delta_f = IGD(c_N).do(p_N)

        # store the delta values to the object
        self.delta_ideal, self.delta_nadir, self.delta_f = delta_ideal, delta_nadir, delta_f

        return max(delta_ideal, delta_nadir, delta_f)


class MultiObjectiveSpaceTerminationWithRenormalization(MultiObjectiveSpaceTermination):

    def __init__(self,
                 n=30,
                 all_to_current=False,
                 sliding_window=True,
                 indicator="igd",
                 **kwargs) -> None:

        super().__init__(**kwargs)
        self.n = n
        self.all_to_current = all_to_current
        self.sliding_window = sliding_window
        self.indicator = indicator

        self.data = []

    def _metric(self, data):
        ret = super()._metric(data)

        if not self.sliding_window:
            data = self.data[-self.metric_window_size:]

        # get necessary data from the current population
        current = data[-1]
        c_F, c_ideal, c_nadir = current["F"], current["ideal"], current["nadir"]

        # normalize all previous generations with respect to current ideal and nadir
        N = [normalize(e["F"], c_ideal, c_nadir) for e in data]

        # check if the movement of all points is significant
        if self.all_to_current:

            c_N = normalize(c_F, c_ideal, c_nadir)
            if self.indicator == "igd":
                delta_f = [IGD(c_N).do(N[k]) for k in range(len(N))]
            elif self.indicator == "hv":
                hv = Hypervolume(ref_point=np.ones(c_F.shape[1]))
                delta_f = [hv.do(N[k]) for k in range(len(N))]
        else:
            delta_f = [IGD(N[k + 1]).do(N[k]) for k in range(len(N) - 1)]

        ret["delta_f"] = delta_f

        return ret

    def _decide(self, metrics):
        delta_ideal = [e["delta_ideal"] for e in metrics]
        delta_nadir = [e["delta_nadir"] for e in metrics]
        delta_f = [max(e["delta_f"]) for e in metrics]
        return max(max(delta_ideal), max(delta_nadir), max(delta_f)) > self.tol



from pymoo.indicators.distance_indicator import DistanceIndicator, euclidean_distance


class GD(DistanceIndicator):

    def __init__(self, pf, **kwargs):
        super().__init__(pf, euclidean_distance, 0, **kwargs)



import numpy as np

from pymoo.core.duplicate import NoDuplicateElimination
from pymoo.core.population import Population
from pymoo.core.repair import NoRepair
from pymoo.util.misc import at_least_2d_array


class Initialization:

    def __init__(self,
                 sampling,
                 repair=None,
                 eliminate_duplicates=None) -> None:

        super().__init__()
        self.sampling = sampling
        self.eliminate_duplicates = eliminate_duplicates if eliminate_duplicates else NoDuplicateElimination()
        self.repair = repair if repair is not None else NoRepair()

    def do(self, problem, n_samples, **kwargs):

        # provide a whole population object - (individuals might be already evaluated)
        if isinstance(self.sampling, Population):
            pop = self.sampling

        else:
            if isinstance(self.sampling, np.ndarray):
                sampling = at_least_2d_array(self.sampling)
                pop = Population.new(X=sampling)
            else:
                pop = self.sampling(problem, n_samples, **kwargs)

        # repair all solutions that are not already evaluated
        not_eval_yet = [k for k in range(len(pop)) if len(pop[k].evaluated) == 0]
        if len(not_eval_yet) > 0:
            pop[not_eval_yet] = self.repair(problem, pop[not_eval_yet], **kwargs)

        # filter duplicate in the population
        pop = self.eliminate_duplicates.do(pop)

        return pop



import numpy as np

from pymoo.core.mutation import Mutation


class ChoiceRandomMutation(Mutation):

    def _do(self, problem, X, **kwargs):
        assert problem.vars is not None

        prob_var = self.get_prob_var(problem, size=len(X))

        for k in range(problem.n_var):
            var = problem.vars[k]
            mut = np.where(np.random.random(len(X)) < prob_var)[0]
            X[mut, k] = var.sample(len(mut))

        return X




import numpy as np


def max_alpha(point, direction, xl, xu, mode="one_hits_bound"):
    bounds = []

    if xl is not None:
        bounds.append(xl)

    if xu is not None:
        bounds.append(xu)

    if len(bounds) == 0:
        return np.inf

    # the bounds in one array
    bounds = np.column_stack(bounds)

    # if the direction is too small we can not divide by 0 - nan will make it being ignored
    dir = direction.copy()
    dir[dir == 0] = np.nan

    # calculate the max factor to be not out of bounds
    val = (bounds - point[:, None]) / dir[:, None]

    # remove nan values by setting them to a negative number
    val[np.isnan(val)] = - np.inf

    # if no value there - no bound exist
    if len(val) == 0:
        return np.inf
    # otherwise return the minimum of values considered
    else:
        if mode == "one_hits_bound":
            if not np.any(val >= 0):
                return 0.0
            else:
                return val[val >= 0].min()
        else:
            return val.max()



import copy

import numpy as np


def default_config():
    return dict(cache=True,
                cv_eps=0.0,
                cv_ieq=dict(scale=None, eps=0.0, pow=None, func=np.sum),
                cv_eq=dict(scale=None, eps=1e-4, pow=None, func=np.sum)
                )


class Individual:
    default_config = default_config

    def __init__(self, config=None, **kwargs) -> None:

        self._X, self._F, self._G, self._H, self._dF, self._dG, self._dH = None, None, None, None, None, None, None
        self._ddF, self._ddG, self._ddH, self._CV, = None, None, None, None
        self.evaluated = None

        # initialize all the local variables
        self.reset()

        # a local storage for data
        self.data = {}

        # the config for this individual
        if config is None:
            config = Individual.default_config()
        self.config = config

        for k, v in kwargs.items():
            if k in self.__dict__:
                self.__dict__[k] = v
            elif "_" + k in self.__dict__:
                self.__dict__["_" + k] = v
            else:
                self.data[k] = v

    def reset(self, data=True):

        empty = np.array([])

        # design variables
        self._X = empty

        # objectives and constraint values
        self._F = empty
        self._G = empty
        self._H = empty

        # first order derivation
        self._dF = empty
        self._dG = empty
        self._dH = empty

        # second order derivation
        self._ddF = empty
        self._ddG = empty
        self._ddH = empty

        # if the constraint violation value to be used
        self._CV = None

        if data:
            self.data = {}

        # a set storing what has been evaluated
        self.evaluated = set()

    def has(self, key):
        return hasattr(self.__class__, key) or key in self.data

    # -------------------------------------------------------
    # Values
    # -------------------------------------------------------

    @property
    def X(self):
        return self._X

    @X.setter
    def X(self, value):
        self._X = value

    @property
    def F(self):
        return self._F

    @F.setter
    def F(self, value):
        self._F = value

    @property
    def G(self):
        return self._G

    @G.setter
    def G(self, value):
        self._G = value

    @property
    def H(self):
        return self._H

    @H.setter
    def H(self, value):
        self._H = value

    @property
    def CV(self):
        config = self.config
        cache = config["cache"]

        if cache and self._CV is not None:
            return self._CV
        else:
            self._CV = np.array([calc_cv(G=self.G, H=self.H, config=config)])
            return self._CV

    @CV.setter
    def CV(self, value):
        self._CV = value

    @property
    def FEAS(self):
        eps = self.config.get("cv_eps", 0.0)
        return self.CV <= eps

    # -------------------------------------------------------
    # Gradients
    # -------------------------------------------------------

    @property
    def dF(self):
        return self._dF

    @dF.setter
    def dF(self, value):
        self._dF = value

    @property
    def dG(self):
        return self._dG

    @dG.setter
    def dG(self, value):
        self._dG = value

    @property
    def dH(self):
        return self._dH

    @dH.setter
    def dH(self, value):
        self._dH = value

    # -------------------------------------------------------
    # Hessians
    # -------------------------------------------------------

    @property
    def ddF(self):
        return self._ddF

    @ddF.setter
    def ddF(self, value):
        self._ddF = value

    @property
    def ddG(self):
        return self._ddG

    @ddG.setter
    def ddG(self, value):
        self._ddG = value

    @property
    def ddH(self):
        return self._ddH

    @ddH.setter
    def ddH(self, value):
        self._ddH = value

    # -------------------------------------------------------
    # Convenience (value instead of array)
    # -------------------------------------------------------

    @property
    def x(self):
        return self.X

    @property
    def f(self):
        return self.F[0]

    @property
    def cv(self):
        if self.CV is None:
            return None
        else:
            return self.CV[0]

    @property
    def feas(self):
        return self.FEAS[0]

    # -------------------------------------------------------
    # Deprecated
    # -------------------------------------------------------

    @property
    def feasible(self):
        return self.FEAS

    # -------------------------------------------------------
    # Other Functions
    # -------------------------------------------------------

    def set_by_dict(self, **kwargs):
        for k, v in kwargs.items():
            self.set(k, v)

    def set(self, key, value):
        if hasattr(self, key):
            setattr(self, key, value)
        else:
            self.data[key] = value
        return self

    def get(self, *keys):
        ret = []

        for key in keys:
            if hasattr(self, key):
                v = getattr(self, key)
            elif key in self.data:
                v = self.data[key]
            else:
                v = None

            ret.append(v)

        if len(ret) == 1:
            return ret[0]
        else:
            return tuple(ret)

    def duplicate(self, key, new_key):
        self.set(new_key, self.get(key))

    def new(self):
        return self.__class__()

    def copy(self, other=None, deep=True):
        obj = self.new()

        # if not provided just copy yourself
        if other is None:
            other = self

        # the data the new object needs to have
        D = other.__dict__

        # if it should be a deep copy do it
        if deep:
            D = copy.deepcopy(D)

        for k, v in D.items():
            obj.__dict__[k] = v

        return obj


def calc_cv(G=None, H=None, config=None):
    if G is None:
        G = np.array([])

    if H is None:
        H = np.array([])

    if config is None:
        config = Individual.default_config()

    if G is None:
        ieq_cv = [0.0]
    elif G.ndim == 1:
        ieq_cv = constr_to_cv(G, **config["cv_ieq"])
    else:
        ieq_cv = [constr_to_cv(g, **config["cv_ieq"]) for g in G]

    if H is None:
        eq_cv = [0.0]
    elif H.ndim == 1:
        eq_cv = constr_to_cv(np.abs(H), **config["cv_eq"])
    else:
        eq_cv = [constr_to_cv(np.abs(h), **config["cv_eq"]) for h in H]

    return np.array(ieq_cv) + np.array(eq_cv)


def constr_to_cv(c, eps=0.0, scale=None, pow=None, func=np.mean):
    if c is None or len(c) == 0:
        return 0.0

    # subtract eps to allow some violation and then zero out all values less than zero
    c = np.maximum(0.0, c - eps)

    # apply init_simplex_scale if necessary
    if scale is not None:
        c = c / scale

    # if a pow factor has been provided
    if pow is not None:
        c = c ** pow

    return func(c)



import numpy as np

from pymoo.algorithms.base.genetic import GeneticAlgorithm
from pymoo.core.survival import Survival
from pymoo.docs import parse_doc_string
from pymoo.operators.crossover.sbx import SBX
from pymoo.operators.mutation.pm import PM
from pymoo.operators.sampling.rnd import FloatRandomSampling
from pymoo.operators.selection.rnd import RandomSelection
from pymoo.termination.max_eval import MaximumFunctionCallTermination
from pymoo.termination.max_gen import MaximumGenerationTermination
from pymoo.util.display.multi import MultiObjectiveOutput
from pymoo.util.misc import has_feasible, vectorized_cdist


class RVEA(GeneticAlgorithm):

    def __init__(self,
                 ref_dirs,
                 alpha=2.0,
                 adapt_freq=0.1,
                 pop_size=None,
                 sampling=FloatRandomSampling(),
                 selection=RandomSelection(),
                 crossover=SBX(eta=30, prob=1.0),
                 mutation=PM(eta=20),
                 eliminate_duplicates=True,
                 n_offsprings=None,
                 output=MultiObjectiveOutput(),
                 **kwargs):
        """

        Parameters
        ----------

        ref_dirs : {ref_dirs}
        adapt_freq : float
            Defines the ratio of generation when the reference directions are updated.
        pop_size : int (default = None)
            By default the population size is set to None which means that it will be equal to the number of reference
            line. However, if desired this can be overwritten by providing a positive number.
        sampling : {sampling}
        selection : {selection}
        crossover : {crossover}
        mutation : {mutation}
        eliminate_duplicates : {eliminate_duplicates}
        n_offsprings : {n_offsprings}

        """

        # set reference directions and pop_size
        self.ref_dirs = ref_dirs
        if self.ref_dirs is not None:
            if pop_size is None:
                pop_size = len(self.ref_dirs)

        # the fraction of n_max_gen when the the reference directions are adapted
        self.adapt_freq = adapt_freq

        # you can override the survival if necessary
        survival = kwargs.pop("survival", None)
        if survival is None:
            survival = APDSurvival(ref_dirs, alpha=alpha)

        super().__init__(pop_size=pop_size,
                         sampling=sampling,
                         selection=selection,
                         crossover=crossover,
                         mutation=mutation,
                         survival=survival,
                         eliminate_duplicates=eliminate_duplicates,
                         n_offsprings=n_offsprings,
                         output=output,
                         **kwargs)

    def _setup(self, problem, **kwargs):

        # if maximum functions termination convert it to generations
        if isinstance(self.termination, MaximumFunctionCallTermination):
            n_gen = np.ceil((self.termination.n_max_evals - self.pop_size) / self.n_offsprings)
            self.termination = MaximumGenerationTermination(n_gen)

        # check whether the n_gen termination is used - otherwise this algorithm can be not run
        if not isinstance(self.termination, MaximumGenerationTermination):
            raise Exception("Please use the n_gen or n_eval as a termination criterion to run RVEA!")

    def _advance(self, **kwargs):
        super()._advance(**kwargs)

        # get the  current generation and maximum of generations
        n_gen, n_max_gen = self.n_gen, self.termination.n_max_gen

        # each i-th generation (define by fr and n_max_gen) the reference directions are updated
        if self.adapt_freq is not None and n_gen % np.ceil(n_max_gen * self.adapt_freq) == 0:
            self.survival.adapt()

    def _set_optimum(self, **kwargs):
        if not has_feasible(self.pop):
            self.opt = self.pop[[np.argmin(self.pop.get("CV"))]]
        else:
            self.opt = self.pop


# ---------------------------------------------------------------------------------------------------------
# Survival Selection
# ---------------------------------------------------------------------------------------------------------

def calc_gamma(V):
    gamma = np.arccos((- np.sort(-1 * V @ V.T))[:, 1])
    gamma = np.maximum(gamma, 1e-64)
    return gamma


def calc_V(ref_dirs):
    return ref_dirs / np.linalg.norm(ref_dirs, axis=1)[:, None]


class APDSurvival(Survival):

    def __init__(self, ref_dirs, alpha=2.0) -> None:
        super().__init__(filter_infeasible=True)
        n_dim = ref_dirs.shape[1]

        self.alpha = alpha
        self.niches = None
        self.V, self.gamma = None, None
        self.ideal, self.nadir = np.full(n_dim, np.inf), None

        self.ref_dirs = ref_dirs
        self.extreme_ref_dirs = np.where(np.any(vectorized_cdist(self.ref_dirs, np.eye(n_dim)) == 0, axis=1))[0]

        self.V = calc_V(self.ref_dirs)
        self.gamma = calc_gamma(self.V)

    def adapt(self):
        self.V = calc_V(calc_V(self.ref_dirs) * (self.nadir - self.ideal))
        self.gamma = calc_gamma(self.V)

    def _do(self, problem, pop, n_survive, algorithm=None, n_gen=None, n_max_gen=None, **kwargs):

        if n_gen is None:
            n_gen = algorithm.n_gen - 1
        if n_max_gen is None:
            n_max_gen = algorithm.termination.n_max_gen

        # get the objective space values
        F = pop.get("F")

        # store the ideal and nadir point estimation for adapt - (and ideal for transformation)
        self.ideal = np.minimum(F.min(axis=0), self.ideal)

        # translate the population to make the ideal point the origin
        F = F - self.ideal

        # the distance to the ideal point
        dist_to_ideal = np.linalg.norm(F, axis=1)
        dist_to_ideal[dist_to_ideal < 1e-64] = 1e-64

        # normalize by distance to ideal
        F_prime = F / dist_to_ideal[:, None]

        # calculate for each solution the acute angles to ref dirs
        acute_angle = np.arccos(F_prime @ self.V.T)
        niches = acute_angle.argmin(axis=1)

        # assign to each reference direction the solution
        niches_to_ind = [[] for _ in range(len(self.V))]
        for k, i in enumerate(niches):
            niches_to_ind[i].append(k)

        # all individuals which will be surviving
        survivors = []

        # for each reference direction
        for k in range(len(self.V)):

            # individuals assigned to the niche
            assigned_to_niche = niches_to_ind[k]

            # if niche not empty
            if len(assigned_to_niche) > 0:

                # the angle of niche to nearest neighboring niche
                gamma = self.gamma[k]

                # the angle from the individuals of this niches to the niche itself
                theta = acute_angle[assigned_to_niche, k]

                # the penalty which is applied for the metric
                M = problem.n_obj if problem.n_obj > 2.0 else 1.0
                penalty = M * ((n_gen / n_max_gen) ** self.alpha) * (theta / gamma)

                # calculate the angle-penalized penalized (APD)
                apd = dist_to_ideal[assigned_to_niche] * (1 + penalty)

                # the individual which survives
                survivor = assigned_to_niche[apd.argmin()]

                # set attributes to the individual
                pop[assigned_to_niche].set(theta=theta, apd=apd, niche=k, opt=False)
                pop[survivor].set("opt", True)

                # select the one with smallest APD value
                survivors.append(survivor)

        ret = pop[survivors]
        self.niches = niches_to_ind
        self.nadir = ret.get("F").max(axis=0)

        return ret


parse_doc_string(RVEA.__init__)



import matplotlib.pyplot as plt

from pymoo.core.callback import Callback
from pymoo.visualization.scatter import Scatter


class AnimationCallback(Callback):

    def __init__(self,
                 do_show=False,
                 do_close=True,
                 nth_gen=1,
                 dpi=None,
                 recorder=None,
                 fname=None,
                 exception_if_not_applicable=True):

        super().__init__()
        self.nth_gen = nth_gen
        self.do_show = do_show
        self.do_close = do_close
        self.exception_if_not_applicable = exception_if_not_applicable

        self.recorder = recorder
        if self.recorder is None and fname is not None:
            try:
                from pyrecorder.recorder import Recorder
                from pyrecorder.writers.video import Video
                from pyrecorder.converters.matplotlib import Matplotlib
                self.recorder = Recorder(Video(fname), converter=Matplotlib(dpi=dpi))
            except:
                raise Exception("Please install or update pyrecorder for animation support: pip install -U pyrecorder")

    def update(self, algorithm):
        if algorithm.n_gen == 1 or algorithm.n_gen % self.nth_gen == 0:
            try:

                figure = self.do(algorithm.problem, algorithm)

                if self.do_show:
                    if figure is not None:
                        figure.show()
                    else:
                        plt.show()

                if self.recorder is not None:
                    self.recorder.record(fig=figure)

                if self.do_close:
                    plt.close(fig=figure)

                return figure

            except Exception as ex:
                if self.exception_if_not_applicable:
                    raise ex

    def do(self, problem, algorithm, **kwargs):
        pass


class ObjectiveSpaceAnimation(AnimationCallback):

    def __init__(self, recorder=None, **kwargs):
        if recorder is None:
            from pyrecorder.recorder import Recorder
            from pyrecorder.writers.streamer import Streamer
            recorder = Recorder(Streamer())
        super().__init__(recorder=recorder, **kwargs)

    def update(self, algorithm):
        F = algorithm.opt.get("F")
        pf = algorithm.problem.pareto_front()

        sc = Scatter()
        sc.add(F)
        if pf is not None:
            sc.add(pf, plot_type="line", color="black", alpha=0.7)
        sc.do()

        self.recorder.record()




import pymoo.gradient.toolbox as anp
import numpy as np

from pymoo.core.problem import Problem
from pymoo.util.remote import Remote


def g_linear(x):
    return 1 + np.sum(x, axis=1)


def g_multimodal(x):
    A = 10
    return 1 + A * x.shape[1] + np.sum(x ** 2 - A * np.cos(2 * np.pi * x), axis=1)


class CTP(Problem):

    def __init__(self, n_var=2, n_ieq_constr=1, option="linear"):
        super().__init__(n_var=n_var, n_obj=2, n_ieq_constr=n_ieq_constr, xl=0, xu=1, vtype=float)

        if option == "linear":
            self.calc_g = g_linear

        elif option == "multimodal":
            self.calc_g = g_multimodal
            self.xl[1:] = -5.12
            self.xu[1:] = 5.12

        else:
            print("Unknown option for CTP single.")

    def calc_objectives(self, x):
        f1 = x[:, 0]
        gg = self.calc_g(x[:, 1:])
        f2 = gg * (1 - (f1 / gg) ** 0.5)
        return f1, f2

    def calc_constraint(self, theta, a, b, c, d, e, f1, f2):

        # Equations in readable format
        exp1 = (f2 - e) * anp.cos(theta) - f1 * anp.sin(theta)

        exp2 = (f2 - e) * anp.sin(theta) + f1 * anp.cos(theta)
        exp2 = b * anp.pi * (exp2 ** c)
        exp2 = anp.abs(anp.sin(exp2))
        exp2 = a * (exp2 ** d)

        # as in the paper
        # val = - (exp1 - exp2)

        # as in the C code of NSGA2
        val = 1 - exp1 / exp2

        # ONE EQUATION
        # _val = - (anp.cos(theta) * (f2 - e) - anp.sin(theta) * f1 -
        #           a * anp.abs(anp.sin(b * anp.pi * (anp.sin(theta) * (f2 - e) + anp.cos(theta) * f1) ** c)) ** d)

        return val

    def _calc_pareto_front(self, *args, **kwargs):
        return Remote.get_instance().load("pymoo", "pf", "CTP", str(self.__class__.__name__).lower() + ".pf")


class CTP1(CTP):

    def __init__(self, n_var=2, n_ieq_constr=2, **kwargs):
        super().__init__(n_var, n_ieq_constr, **kwargs)

        a, b = np.zeros(n_ieq_constr + 1), np.zeros(n_ieq_constr + 1)
        a[0], b[0] = 1, 1
        delta = 1 / (n_ieq_constr + 1)
        alpha = delta

        for j in range(n_ieq_constr):
            beta = a[j] * np.exp(-b[j] * alpha)
            a[j + 1] = (a[j] + beta) / 2
            b[j + 1] = - 1 / alpha * np.log(beta / a[j + 1])

            alpha += delta

        self.a = a[1:]
        self.b = b[1:]

    def _evaluate(self, x, out, *args, **kwargs):
        f1 = x[:, 0]
        gg = self.calc_g(x[:, 1:])
        f2 = gg * anp.exp(-f1 / gg)
        out["F"] = anp.column_stack([f1, f2])

        a, b = self.a, self.b
        g = []
        for j in range(self.n_ieq_constr):
            _g = - (f2 - (a[j] * anp.exp(-b[j] * f1)))
            g.append(_g)
        out["G"] = anp.column_stack(g)


class CTP2(CTP):

    def _evaluate(self, x, out, *args, **kwargs):
        f1, f2 = self.calc_objectives(x)
        out["F"] = anp.column_stack([f1, f2])

        theta = -0.2 * anp.pi
        a, b, c, d, e = 0.2, 10, 1, 6, 1
        out["G"] = self.calc_constraint(theta, a, b, c, d, e, f1, f2)


class CTP3(CTP):

    def _evaluate(self, x, out, *args, **kwargs):
        f1, f2 = self.calc_objectives(x)
        out["F"] = anp.column_stack([f1, f2])

        theta = -0.2 * anp.pi
        a, b, c, d, e = 0.1, 10, 1, 0.5, 1

        out["G"] = self.calc_constraint(theta, a, b, c, d, e, f1, f2)


class CTP4(CTP):

    def _evaluate(self, x, out, *args, **kwargs):
        f1, f2 = self.calc_objectives(x)
        out["F"] = anp.column_stack([f1, f2])

        theta = -0.2 * anp.pi
        a, b, c, d, e = 0.75, 10, 1, 0.5, 1

        out["G"] = self.calc_constraint(theta, a, b, c, d, e, f1, f2)


class CTP5(CTP):

    def _evaluate(self, x, out, *args, **kwargs):
        f1, f2 = self.calc_objectives(x)
        out["F"] = anp.column_stack([f1, f2])

        theta = -0.2 * anp.pi
        a, b, c, d, e = 0.1, 10, 2, 0.5, 1

        out["G"] = self.calc_constraint(theta, a, b, c, d, e, f1, f2)


class CTP6(CTP):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.xu = np.full(self.n_var, 20)
        self.xu[0] = 1

    def _evaluate(self, x, out, *args, **kwargs):
        f1, f2 = self.calc_objectives(x)
        out["F"] = anp.column_stack([f1, f2])

        theta = 0.1 * anp.pi
        a, b, c, d, e = 40, 0.5, 1, 2, -2

        out["G"] = self.calc_constraint(theta, a, b, c, d, e, f1, f2)


class CTP7(CTP):

    def _evaluate(self, x, out, *args, **kwargs):
        f1, f2 = self.calc_objectives(x)
        out["F"] = anp.column_stack([f1, f2])

        theta = -0.05 * anp.pi
        a, b, c, d, e = 40, 5, 1, 6, 0

        out["G"] = self.calc_constraint(theta, a, b, c, d, e, f1, f2)


class CTP8(CTP):
    def __init__(self, **kwargs):
        super().__init__(n_ieq_constr=2, **kwargs)
        self.xu = np.full(self.n_var, 20)
        self.xu[0] = 1

    def _evaluate(self, x, out, *args, **kwargs):
        f1, f2 = self.calc_objectives(x)
        out["F"] = anp.column_stack([f1, f2])

        theta = 0.1 * anp.pi
        a, b, c, d, e = 40, 0.5, 1, 2, -2
        g1 = self.calc_constraint(theta, a, b, c, d, e, f1, f2)

        theta = -0.05 * anp.pi
        a, b, c, d, e = 40, 2, 1, 6, 0
        g2 = self.calc_constraint(theta, a, b, c, d, e, f1, f2)

        out["G"] = anp.column_stack([g1, g2])


if __name__ == '__main__':
    problem = CTP1(n_ieq_constr=3)
    print(problem.n_ieq_constr)



import numpy as np

from pymoo.algorithms.moo.nsga3 import NSGA3
from pymoo.operators.selection.tournament import compare, TournamentSelection


# =========================================================================================================
# Implementation
# =========================================================================================================


def comp_by_rank_and_ref_line_dist(pop, P, **kwargs):
    S = np.full(P.shape[0], np.nan)

    for i in range(P.shape[0]):
        a, b = P[i, 0], P[i, 1]

        # if at least one solution is infeasible
        if pop[a].CV > 0.0 or pop[b].CV > 0.0:
            S[i] = compare(a, pop[a].CV, b, pop[b].CV, method='smaller_is_better', return_random_if_equal=True)

        # both solutions are feasible
        else:

            # if in the same niche select by rank
            if pop[a].get("niche") == pop[b].get("niche"):

                if pop[a].get("rank") != pop[b].get("rank"):
                    S[i] = compare(a, pop[a].get("rank"), b, pop[b].get("rank"), method='smaller_is_better')

                else:
                    S[i] = compare(a, pop[a].get("dist_to_niche"), b, pop[b].get("dist_to_niche"),
                                   method='smaller_is_better')

        if np.isnan(S[i]):
            S[i] = np.random.choice([a, b])

    return S[:, None].astype(int)


class UNSGA3(NSGA3):

    def __init__(self,
                 ref_dirs,
                 selection=TournamentSelection(func_comp=comp_by_rank_and_ref_line_dist),
                 **kwargs):
        super().__init__(ref_dirs, selection=selection, **kwargs)



import numpy as np


class SlidingWindow(list):

    def __init__(self, size=None) -> None:
        super().__init__()
        self.size = size

    def append(self, entry):
        super().append(entry)

        if self.size is not None:
            while len(self) > self.size:
                self.pop(0)

    def is_full(self):
        return self.size == len(self)

    def to_numpy(self):
        return np.array(self)

    def clear(self):
        while len(self) > 0:
            self.pop()



import numpy as np

from pymoo.core.decision_making import DecisionMaking, find_outliers_upper_tail, NeighborFinder


class HighTradeoffPoints(DecisionMaking):

    def __init__(self, epsilon=0.125, **kwargs) -> None:
        super().__init__(**kwargs)
        self.epsilon = epsilon

    def _do(self, F, **kwargs):
        n, m = F.shape

        neighbors_finder = NeighborFinder(F, epsilon=0.125, n_min_neigbors="auto", consider_2d=False)

        mu = np.full(n, - np.inf)

        # for each solution in the set calculate the least amount of improvement per unit deterioration
        for i in range(n):

            # for each neighbour in a specific radius of that solution
            neighbors = neighbors_finder.find(i)

            # calculate the trade-off to all neighbours
            diff = F[neighbors] - F[i]

            # calculate sacrifice and gain
            sacrifice = np.maximum(0, diff).sum(axis=1)
            gain = np.maximum(0, -diff).sum(axis=1)

            np.warnings.filterwarnings('ignore')
            tradeoff = sacrifice / gain

            # otherwise find the one with the smalled one
            mu[i] = np.nanmin(tradeoff)

        return find_outliers_upper_tail(mu)



import numpy as np

from pymoo.core.problem import Problem
from pymoo.operators.sampling.lhs import LatinHypercubeSampling
from pymoo.util.reference_direction import ReferenceDirectionFactory, map_onto_unit_simplex


class RandomSamplingAndMap(ReferenceDirectionFactory):

    def __init__(self,
                 n_dim,
                 n_points,
                 **kwargs):
        super().__init__(n_dim, **kwargs)
        self.n_points = n_points

    def _do(self):
        problem = Problem(n_var=self.n_dim, xl=0.0, xu=1.0)
        sampling = LatinHypercubeSampling()

        x = sampling(problem, self.n_points - self.n_dim, to_numpy=True)
        x = map_onto_unit_simplex(x, "kraemer")
        x = np.row_stack([x, np.eye(self.n_dim)])
        return x



from pymoo.core.individual import calc_cv
from pymoo.core.meta import Meta
from pymoo.core.problem import Problem
from pymoo.util.misc import from_dict

import numpy as np


class ConstraintsAsPenalty(Meta, Problem):

    def __init__(self,
                 problem,
                 penalty: float = 0.1):
        super().__init__(problem)

        # the amount of penalty to add for this type
        self.penalty = penalty

        # set ieq and eq to zero (because it became now a penalty)
        self.n_ieq_constr = 0
        self.n_eq_constr = 0

    def do(self, X, return_values_of, *args, **kwargs):
        out = self.__object__.do(X, return_values_of, *args, **kwargs)

        # get at the values from the output
        F, G, H = from_dict(out, "F", "G", "H")

        # store a backup of the values in out
        out["__F__"], out["__G__"], out["__H__"] = F, G, H

        # calculate the total constraint violation (here normalization shall be already included)
        CV = calc_cv(G=G, H=H)

        # set the penalized objective values
        out["F"] = F + self.penalty * np.reshape(CV, F.shape)

        del out["G"]
        del out["H"]

        return out



import numpy as np

from pymoo.core.problem import ElementwiseProblem
from pymoo.util.remote import Remote


class MODAct(ElementwiseProblem):
    """Multi-Objective Design of Actuators

    MODAct is a framework for real-world constrained multi-objective optimization.
    Refer to the python package https://github.com/epfl-lamd/modact from requirements.

    Best-known Pareto fronts must be downloaded from here: https://doi.org/10.5281/zenodo.3824302

    Parameters
    ----------

    function: str or modact.problems
        The name of the benchmark problem to use either as a string or the
        problem object instance. Example values: cs1, cs3, ct2, ct4, cts3

    References:
    ----------
    C. Picard and J. Schiffmann, “Realistic Constrained Multi-Objective Optimization Benchmark Problems from Design,”
    IEEE Transactions on Evolutionary Computation, pp. 1–1, 2020.
    """

    def __init__(self, function, pf=None, **kwargs):

        self.function = function
        self.pf = pf

        try:
            import modact.problems as pb
        except:
            raise Exception("Please install the modact library: https://github.com/epfl-lamd/modact")

        if isinstance(function, pb.Problem):
            self.fct = function
        else:
            self.fct = pb.get_problem(function)

        lb, ub = self.fct.bounds()
        n_var = len(lb)
        n_obj = len(self.fct.weights)
        n_ieq_constr = len(self.fct.c_weights)
        xl = lb
        xu = ub

        self.weights = np.array(self.fct.weights)
        self.c_weights = np.array(self.fct.c_weights)

        super().__init__(n_var=n_var, n_obj=n_obj, n_ieq_constr=n_ieq_constr, xl=xl, xu=xu, vtype=float, **kwargs)

    def _evaluate(self, x, out, *args, **kwargs):
        f, g = self.fct(x)
        out["F"] = np.array(f) * -1 * self.weights
        out["G"] = np.array(g) * self.c_weights

    def _calc_pareto_front(self, *args, **kwargs):
        # allows to provide a custom pf - because of the size of files published by the author
        if self.pf is None:
            pf = Remote.get_instance().load("pymoo", "pf", "MODACT", f"{self.function}.pf")
            # pf = pf * [1, -1]
            pf = pf * self.weights * -1
            return pf
        else:
            return self.pf



import numpy as np

from pymoo.core.crossover import Crossover


def remove_from_adj_list(H, val):
    for e in list(H[val]):
        H[e].remove(val)
    del H[val]


def has_duplicates(x):
    H = set()
    for v in x:
        if v in H:
            return True
        H.add(v)
    return False


def erx(a, b):
    """
    http://www.rubicite.com/Tutorials/GeneticAlgorithms/CrossoverOperators/EdgeRecombinationCrossoverOperator.aspx

    Algorithm Pseudo Code:

    1. X = the first node from a random parent.

    2. While the CHILD chromo isn't full, Loop:
        - Append X to CHILD
        - Remove X from Neighbor Lists

        if X's neighbor list is empty:
           - Xp = random node not already in CHILD
        else
           - Determine neighbor of X that has fewest neighbors
           - If there is a tie, randomly choose 1
           - Xp = chosen node
        X = Xp
     """

    assert len(a) == len(b)

    # calculate the edge matrix considering both permutation
    H = calc_adjency_matrix(a)
    H = calc_adjency_matrix(b, H=H)

    # randomly select the first node
    _next = np.random.choice(list(H.keys()))

    y = []
    while True:

        # append to the child
        y.append(_next)

        # break if the child was successfully created.
        if len(y) == len(a):
            break

        # get the neighbors to consider and remove them from the lists
        neighbors = list(H[_next])
        remove_from_adj_list(H, _next)

        # if the current node does not have any neighbors
        if len(neighbors) == 0:
            _next = np.random.choice(list(H.keys()))

        # otherwise search in the neighbors for a node with the fewest neighbors
        else:
            # search for the one with minimum neighbors
            n_neighbors = [len(H[e]) for e in neighbors]
            min_n_neighbors = min(n_neighbors)
            _next = [neighbors[k] for k in range(len(neighbors)) if n_neighbors[k] == min_n_neighbors]

            # break the tie if they might have the same number of neighbors
            _next = np.random.choice(_next)

    return y


class EdgeRecombinationCrossover(Crossover):

    def __init__(self, **kwargs):
        super().__init__(2, 1, **kwargs)

    def _do(self, problem, X, **kwargs):
        _, n_matings, n_var = X.shape
        Y = np.full((self.n_offsprings, n_matings, n_var), -1, dtype=int)

        for i in range(n_matings):
            a, b = X[:, i, :]
            Y[0, i, :] = erx(a, b)

        return Y

class ERX(EdgeRecombinationCrossover):
    pass

def number_to_letter(n):
    return chr(ord('@') + n)


def numbers_to_letters(numbers):
    return [number_to_letter(n) for n in numbers]


def letter_to_number(char):
    return ord(char.lower()) - 96


def letters_to_numbers(letters):
    return np.array([letter_to_number(char) for char in letters])


def calc_adjency_matrix(x, H=None):
    H = {} if H is None else H

    for k in range(len(x)):
        prev = (k - 1) % len(x)
        succ = (k + 1) % len(x)

        if x[k] not in H:
            H[x[k]] = set()
        H[x[k]].update([x[prev], x[succ]])

    return H


if __name__ == "__main__":
    a = ['A', 'B', 'F', 'E', 'D', 'G', 'C']
    b = ['G', 'F', 'A', 'B', 'C', 'D', 'E']

    """
    A: B C F
    B: A F C
    C: A G B D
    D: E G C E
    E: F D G
    F: B E G A
    G: D C E F
    """

    H = calc_adjency_matrix(a)
    H = calc_adjency_matrix(b, H=H)

    assert len(H["A"]) == 3
    assert 'B' in H["A"] and 'C' in H["A"] and 'F' in H["A"]
    assert len(H["B"]) == 3
    assert 'A' in H["B"] and 'F' in H["B"] and 'C' in H["B"]
    assert len(H["C"]) == 4
    assert 'A' in H["C"] and 'G' in H["C"] and 'B' in H["C"] and 'D' in H["C"]
    assert len(H["D"]) == 3
    assert 'E' in H["D"] and 'G' in H["D"] and 'C' in H["D"]
    assert len(H["E"]) == 3
    assert 'F' in H["E"] and 'D' in H["E"] and 'G' in H["E"]
    assert len(H["F"]) == 4
    assert 'B' in H["F"] and 'E' in H["F"] and 'G' in H["F"] and 'A' in H["F"]
    assert len(H["G"]) == 4
    assert 'D' in H["G"] and 'E' in H["G"] and 'E' in H["G"] and 'F' in H["G"]

    c = erx(a, b)
